# test_sampling.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2009 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""Tests for the katcp.sampling module.
   """

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import concurrent.futures
import logging
import threading

import tornado.testing

from _thread import get_ident
from tornado import gen

from katcp import Sensor, sampling
from katcp.testutils import (DeviceTestSensor, TestLogHandler,
                             TimewarpAsyncTestCase)

log_handler = TestLogHandler()
logging.getLogger("katcp").addHandler(log_handler)
logger = logging.getLogger(__name__)


class TestSampling(TimewarpAsyncTestCase):
    # TODO Also test explicit ioloop passing

    def setUp(self):
        """Set up for test."""
        # test sensor
        super(TestSampling, self).setUp()
        self.sensor = DeviceTestSensor(
                Sensor.INTEGER, "an.int", "An integer.", "count",
                [-40, 30],
                timestamp=self.ioloop_time, status=Sensor.NOMINAL, value=3)
        # test callback
        def inform(sensor, reading):
            assert get_ident() == self.ioloop_thread_id, (
                "inform must be called from in the ioloop")
            self.calls.append((sensor, reading))

        self.calls = []
        self.inform = inform

    def test_sampling(self):
        """Test getting and setting the sampling."""
        s = self.sensor

        sampling.SampleNone(None, s)
        sampling.SampleAuto(None, s)
        sampling.SamplePeriod(None, s, 10)
        sampling.SampleEvent(None, s)
        sampling.SampleDifferential(None, s, 2)
        self.assertRaises(ValueError, sampling.SampleNone, None, s, "foo")
        self.assertRaises(ValueError, sampling.SampleAuto, None, s, "bar")
        self.assertRaises(ValueError, sampling.SamplePeriod, None, s)
        self.assertRaises(ValueError, sampling.SamplePeriod, None, s, "0")
        self.assertRaises(ValueError, sampling.SamplePeriod, None, s, "-1")
        self.assertRaises(ValueError, sampling.SampleEvent, None, s, "foo")
        self.assertRaises(ValueError, sampling.SampleDifferential, None, s)
        self.assertRaises(ValueError, sampling.SampleDifferential,
                          None, s, "-1")
        self.assertRaises(ValueError, sampling.SampleDifferential,
                          None, s, "1.5")

        sampling.SampleStrategy.get_strategy("none", None, s)
        sampling.SampleStrategy.get_strategy("auto", None, s)
        sampling.SampleStrategy.get_strategy("period", None, s, "15")
        sampling.SampleStrategy.get_strategy("event", None, s)
        sampling.SampleStrategy.get_strategy("differential", None, s, "2")
        sampling.SampleStrategy.get_strategy(b"none", None, s)
        sampling.SampleStrategy.get_strategy(b"auto", None, s)
        sampling.SampleStrategy.get_strategy(b"period", None, s, "15")
        sampling.SampleStrategy.get_strategy(b"event", None, s)
        sampling.SampleStrategy.get_strategy(b"differential", None, s, "2")
        self.assertRaises(ValueError, sampling.SampleStrategy.get_strategy,
                          "random", None, s)
        self.assertRaises(ValueError, sampling.SampleStrategy.get_strategy,
                          "period", None, s, "foo")
        self.assertRaises(ValueError, sampling.SampleStrategy.get_strategy,
                          "differential", None, s, "bar")

    @tornado.testing.gen_test(timeout=200)
    # Timeout needs to be longer than 'fake' duration of the test, since the tornado
    # ioloop is using out time-warped clock to determine timeouts too!
    def test_periodic(self):
        t0 = self.ioloop_time
        sample_p = 10                            # sample DUT in seconds
        DUT = sampling.SamplePeriod(self.inform, self.sensor, sample_p)
        self.assertEqual(DUT.get_sampling_formatted(), (b'period', [b'10']))
        self.assertEqual(self.calls, [])
        t, status, value = self.sensor.read()
        DUT.start()
        yield self.wake_ioloop()
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []
        # Warp the ioloop clock forward a bit more than one DUT. Check that
        #   1) a sample is sent,
        #   2) the next sample is scheduled at t0+DUT, not t0+DUT+extra delay
        yield self.set_ioloop_time(t0 + sample_p*1.15)
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []
        # Don't expect an update, since we are at just before the next sample DUT
        yield self.set_ioloop_time(t0 + sample_p*1.99)
        self.assertEqual(self.calls, [])
        # Now we are at exactly the next sample time, expect update
        yield self.set_ioloop_time(t0 + sample_p*2)
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []
        # Bit past previous sample time, expect no update
        yield self.set_ioloop_time(t0 + sample_p*2.16)
        self.assertEqual(self.calls, [])

        # Check that no update is sent if the sensor is updated, but that the next
        # periodic update is correct
        t, status, value = (t0 + sample_p*2.5, Sensor.WARN, -1)
        self.sensor.set(t, status, value)
        yield self.set_ioloop_time(t0 + sample_p*2.6)
        self.assertEqual(self.calls, [])
        yield self.set_ioloop_time(t0 + sample_p*3)
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []

        # Cancel strategy and check that its timeout call is cancelled.
        DUT.cancel()
        yield self.wake_ioloop()
        yield self.set_ioloop_time(t0 + sample_p*4.1)
        self.assertEqual(self.calls, [])


    @tornado.testing.gen_test(timeout=200)
    def test_auto(self):
        t0 = self.ioloop_time
        DUT = sampling.SampleAuto(self.inform, self.sensor)
        self.assertEqual(DUT.get_sampling_formatted(), (b'auto', []))
        self.assertEqual(self.calls, [])
        t, status, value = self.sensor.read()
        DUT.start()
        yield self.wake_ioloop()
        # Check that it is attached
        self.assertTrue(DUT in self.sensor._observers)
        # The initial update
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []
        # Move along in time, don't expect any updates
        yield self.set_ioloop_time(t0 + 20)
        self.assertEqual(self.calls, [])
        # Now update the sensor a couple of times
        t1, status1, value1 = t0 + 21, Sensor.ERROR, 2
        t2, status2, value2 = t0 + 22, Sensor.NOMINAL, -1
        self.sensor.set(t1, status1, value1)
        self.sensor.set(t2, status2, value2)
        self.assertEqual(self.calls, [(self.sensor, (t1, status1, value1)),
                                      (self.sensor, (t2, status2, value2))])
        self.calls = []

        self._thread_update_check(t, status, value)
        yield self._check_cancel(DUT)

    @gen.coroutine
    def _thread_update_check(self, ts, status, value):
        # Check update from thread (inform() raises if called from the wrong thread)
        # Clears out self.calls before starting
        self.calls = []
        f = concurrent.futures.Future()
        def do_update():
            try:
                self.sensor.set(ts, status, value)
            finally:
                f.set_result(None)
            return f
        t = threading.Thread(target=do_update)
        t.start()
        yield f
        yield self.wake_ioloop()
        self.assertEqual(self.calls, [(self.sensor, (ts, status, value))])

    @gen.coroutine
    def _check_cancel(self, DUT):
        # Check post-cancel cleanup
        DUT.cancel()
        yield self.wake_ioloop()
        self.assertFalse(DUT in self.sensor._observers)

    @tornado.testing.gen_test(timeout=200)
    def test_differential(self):
        """Test SampleDifferential strategy."""
        t, status, value = self.sensor.read()
        delta = 3
        DUT = sampling.SampleDifferential(self.inform, self.sensor, delta)
        self.assertEqual(DUT.get_sampling_formatted(), (b'differential', [b'3']))
        self.assertEqual(len(self.calls), 0)
        DUT.start()
        yield self.wake_ioloop()
        # Check initial update
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []
        # Some Updates less than delta from initial value
        self.sensor.set_value(value + 1)
        self.sensor.set_value(value + delta)
        self.sensor.set_value(value)
        self.sensor.set_value(value - 2)
        self.assertEqual(len(self.calls), 0)
        # Now an update bigger than delta from initial value
        self.sensor.set(t, status, value + delta + 1)
        yield self.wake_ioloop()
        self.assertEqual(self.calls, [(self.sensor, (t, status, value + delta + 1))])
        self.calls = []
        # Now change only the status, should update
        t, status, value = self.sensor.read()
        self.sensor.set(t, Sensor.ERROR, value)
        self.assertEqual(self.calls, [(self.sensor, (t, Sensor.ERROR, value))])
        # Test threaded update
        yield self._thread_update_check(t, status, value)
        yield self._check_cancel(DUT)

    def test_differential_timestamp(self):
        # Test that the timetamp differential is stored correctly as
        # seconds. This is mainly to check the conversion of the katcp spec from
        # milliseconds to seconds for katcp v5 spec.
        time_diff = 4.12                  # Time differential in seconds
        ts_sensor = Sensor(Sensor.TIMESTAMP, 'ts', 'ts sensor', '')
        diff = sampling.SampleDifferential(self.inform, ts_sensor, time_diff)
        self.assertEqual(diff.get_sampling_formatted(), (b'differential', [b'4.12']))
        self.assertEqual(diff._threshold, time_diff)

    @tornado.testing.gen_test(timeout=200)
    def test_event_rate(self):
        """Test SampleEventRate strategy."""
        shortest = 1.5
        longest = 4.5
        t, status, value = self.sensor.read()

        DUT = sampling.SampleEventRate(self.inform, self.sensor, shortest, longest)
        self.assertEqual(DUT.get_sampling_formatted(), (b'event-rate', [b'1.5', b'4.5']))
        self.assertEqual(len(self.calls), 0)
        DUT.start()
        yield self.wake_ioloop()
        # Check initial update
        t_last_sent = self.ioloop_time
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []

        # Too soon, should not send update
        yield self.set_ioloop_time(t_last_sent + shortest*0.99)
        value = value + 3
        t = self.ioloop_time
        self.sensor.set(t, status, value)
        self.assertEqual(len(self.calls), 0)

        # Too soon again, should not send update
        yield self.set_ioloop_time(t_last_sent + shortest*0.999)
        value = value + 1
        t = self.ioloop_time
        self.sensor.set(t, status, value)
        self.assertEqual(len(self.calls), 0)

        # Should now get minimum time update
        yield self.set_ioloop_time(t_last_sent + shortest)
        t_last_sent = self.ioloop_time
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []

        # Warp to just before longest period, should not update
        yield self.set_ioloop_time(t_last_sent + longest*0.999)
        self.assertEqual(len(self.calls), 0)

        # Warp to longest period, should update
        yield self.set_ioloop_time(t_last_sent + longest)
        t_last_sent = self.ioloop_time
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []

        # Warp to just before next longest period, should not update
        yield self.set_ioloop_time(t_last_sent + longest*0.999)
        self.assertEqual(len(self.calls), 0)

        # Warp to longest period, should update
        yield self.set_ioloop_time(t_last_sent + longest)
        t_last_sent = self.ioloop_time
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []

        # Set identical value, jump past min update time, no update should happen
        self.sensor.set(self.ioloop_time, status, value)
        yield self.set_ioloop_time(t_last_sent + shortest)
        self.assertEqual(len(self.calls), 0)

        # Set new value, update should happen
        value = value - 2
        self.sensor.set(self.ioloop_time, status, value)
        t_last_sent = self.ioloop_time
        self.assertEqual(self.calls, [(self.sensor, (t_last_sent, status, value))])
        self.calls = []

        # Now warp to after min period, change only status, update should happen
        yield self.set_ioloop_time(t_last_sent + shortest)
        status = Sensor.ERROR
        self.sensor.set(self.ioloop_time, status, value)
        self.assertEqual(self.calls, [(self.sensor, (self.ioloop_time, status, value))])
        t_last_sent = self.ioloop_time
        self.calls = []

        yield self.set_ioloop_time(t_last_sent + shortest)
        status = Sensor.NOMINAL
        value = value + 1
        yield self._thread_update_check(self.ioloop_time, status, value)
        yield self._check_cancel(DUT)
        self.calls = []

        # Since strategy is cancelled, no futher updates should be sent
        yield self.set_ioloop_time(self.ioloop_time + 5*longest)
        self.sensor.set(self.ioloop_time, Sensor.WARN, value + 3)
        self.assertEqual(len(self.calls), 0)

    @tornado.testing.gen_test(timeout=2000000)
    def test_event(self):
        """Test SampleEvent strategy."""
        DUT = sampling.SampleEvent(self.inform, self.sensor)
        self.assertEqual(DUT.get_sampling_formatted(), (b'event', []))
        self.assertEqual(self.calls, [])
        DUT.start()
        yield self.wake_ioloop()
        # Check initial update
        self.assertEqual(len(self.calls), 1)

        # Jump forward a lot, should not result in another sample
        yield self.set_ioloop_time(200000)
        self.assertEqual(len(self.calls), 1)

        self.sensor.set_value(2, status=Sensor.NOMINAL)
        self.assertEqual(len(self.calls), 2)

        # Test that an update is suppressed if the sensor value is unchanged
        self.sensor.set_value(2, status=Sensor.NOMINAL)
        self.assertEqual(len(self.calls), 2)

        # Test that an update happens if the status changes even if the value is
        # unchanged
        self.sensor.set_value(2, status=Sensor.WARN)
        self.assertEqual(len(self.calls), 3)

    @tornado.testing.gen_test(timeout=200)
    def test_differential_rate(self):
        delta = 2
        shortest = 1.5
        longest = 4.5
        t, status, value = self.sensor.read()

        DUT = sampling.SampleDifferentialRate(
            self.inform, self.sensor, delta, shortest, longest)
        self.assertEqual(
            DUT.get_sampling_formatted(), (b'differential-rate', [b'2', b'1.5', b'4.5']))
        self.assertEqual(len(self.calls), 0)
        DUT.start()
        yield self.wake_ioloop()
        # Check initial update
        t_last_sent = self.ioloop_time
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []

        # Too soon, should not send update
        yield self.set_ioloop_time(t_last_sent + shortest*0.99)
        value = value + delta + 1
        t = self.ioloop_time
        self.sensor.set(t, status, value)
        self.assertEqual(len(self.calls), 0)

        # Too soon again, should not send update
        yield self.set_ioloop_time(t_last_sent + shortest*0.999)
        value = value + delta + 1
        t = self.ioloop_time
        self.sensor.set(t, status, value)
        self.assertEqual(len(self.calls), 0)

        # Should now get minimum time update
        yield self.set_ioloop_time(t_last_sent + shortest)
        t_last_sent = self.ioloop_time
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []

        # Warp to just before longest period, should not update
        yield self.set_ioloop_time(t_last_sent + longest*0.999)
        self.assertEqual(len(self.calls), 0)

        # Warp to longest period, should update
        yield self.set_ioloop_time(t_last_sent + longest)
        t_last_sent = self.ioloop_time
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []

        # Warp to just before next longest period, should not update
        yield self.set_ioloop_time(t_last_sent + longest*0.999)
        self.assertEqual(len(self.calls), 0)

        # Warp to longest period, should update
        yield self.set_ioloop_time(t_last_sent + longest)
        t_last_sent = self.ioloop_time
        self.assertEqual(self.calls, [(self.sensor, (t, status, value))])
        self.calls = []

        # Set value with to small a change, jump past min update time, no update should
        # happen
        value = value - delta
        self.sensor.set(self.ioloop_time, status, value)
        yield self.set_ioloop_time(t_last_sent + shortest)
        self.assertEqual(len(self.calls), 0)

        # Set new value with large enough difference, update should happen
        value = value - 1
        self.sensor.set(self.ioloop_time, status, value)
        t_last_sent = self.ioloop_time
        self.assertEqual(self.calls, [(self.sensor, (t_last_sent, status, value))])
        self.calls = []

        # Now warp to after min period, change only status, update should happen
        yield self.set_ioloop_time(t_last_sent + shortest)
        status = Sensor.ERROR
        self.sensor.set(self.ioloop_time, status, value)
        self.assertEqual(self.calls, [(self.sensor, (self.ioloop_time, status, value))])
        t_last_sent = self.ioloop_time
        self.calls = []

        yield self.set_ioloop_time(t_last_sent + shortest)
        status = Sensor.NOMINAL
        value = value + 1
        yield self._thread_update_check(self.ioloop_time, status, value)
        yield self._check_cancel(DUT)
        self.calls = []

        # Since strategy is cancelled, no futher updates should be sent
        yield self.set_ioloop_time(self.ioloop_time + 5*longest)
        self.sensor.set(self.ioloop_time, Sensor.WARN, value + 3)
        self.assertEqual(len(self.calls), 0)



# Copyright 2008 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""The empty module is here so that pylint can examine the scripts

   @author Robert Crida <robert.crida@ska.ac.za>
   @date 2008-10-10
   """

# Automatically added by katversion
__version__ = '0.9.0'



# Copyright 2014 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function

from tornado.util import ObjectDict

from katcp import Sensor
import tornado_io

logging.basicConfig(
    format="%(asctime)s %(name)s %(levelname)s %(funcName)s(%(filename)s:%(lineno)d)%(message)s",
    level=logging.INFO
)

class D(DeviceServer):
    def setup_sensors(self):
        self.add_sensor(Sensor.boolean('asens'))

    def request_stupid(self, req, msg):
        "blah"
        req.inform('Doing stupid sleep')
        time.sleep(10)
        req.inform('Stupid sleep done')
        return Message.reply('stupid', 'ok')

    def start(self, timeout=None):
        """Start the server in a new thread.

        Parameters
        ----------
        timeout : float in seconds
            Time to wait for server thread to start.
        """
        super(D, self).start(timeout)
        s = self.get_sensor('asens')
        def updat():
            s.set_value(not s.value())
            self.ioloop.call_later(0.9, updat)
        self.ioloop.add_callback(updat)

restart_queue = Queue.Queue()
DS = D('', 5000)
DS.set_restart_queue(restart_queue)
try:
    DS.start()
    import IPython ; IPython.embed()
    # while True:
    #     try:
    #         restart_device = restart_queue.get(timeout=100000)
    #     except Queue.Empty:
    #         continue
    #     logging.info('Stopping')
    #     restart_device.stop()
    #     restart_device.join()
    #     logging.info("Restarting ...")
    #     restart_device.start()
    #     logging.info("Started.")

except Exception:
    logging.exception("error")
finally:
    DS.stop()



# client.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2009 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

"""Clients for the KAT device control language."""

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import logging
import sys
import traceback

from builtins import object
from concurrent.futures import Future, TimeoutError
from functools import partial, wraps

import tornado.ioloop
import tornado.iostream
import tornado.tcpclient

from _thread import get_ident as get_thread_ident
from future.utils import with_metaclass
from tornado import gen
from tornado.concurrent import Future as tornado_Future
from tornado.util import ObjectDict

from .core import (FLOAT_TS_KATCP_MAJOR, SEC_TO_MS_FAC, SEC_TS_KATCP_MAJOR,
                   AsyncEvent, DeviceMetaclass, KatcpClientDisconnected,
                   KatcpClientError, KatcpVersionError, LatencyTimer, Message,
                   MessageParser, ProtocolFlags, ensure_native_str, until_later)
from .ioloop_manager import IOLoopManager

# logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger("katcp.client")

def address_to_string(addr_tuple):
    return ":".join(str(part) for part in addr_tuple)


def make_threadsafe(meth):
    """Decorator for a DeviceClient method that should always run in ioloop.

    Used with DeviceClient.enable_thread_safety(). If not called the method
    will be unprotected and it is the user's responsibility to ensure that
    these methods are only called from the ioloop, otherwise the decorated
    methods are wrapped. Should only be used for functions that have no return
    value.

    """
    meth.make_threadsafe = True
    return meth


def make_threadsafe_blocking(meth):
    """Decorator for a DeviceClient method that will block.

    Used with DeviceClient.enable_thread_safety(). Used to provide blocking
    calls that can be made from other threads. If called in ioloop context,
    calls the original method directly to prevent deadlocks. Will route return
    value to caller. Add `timeout` keyword argument to limit blocking time.
    If `meth` returns a future, its result will be returned, otherwise its
    result will be passed back directly.

    """
    meth.make_threadsafe_blocking = True
    return meth


class DeviceClient(with_metaclass(DeviceMetaclass, object)):
    """Device client proxy.

    Subclasses should implement .reply\_*, .inform\_* and
    send_request\_* methods to take actions when messages arrive,
    and implement unhandled_inform, unhandled_reply and
    unhandled_request to provide fallbacks for messages for
    which there is no handler.

    Request messages can be sent by calling .send_request().

    Parameters
    ----------
    host : string
        Host to connect to.
    port : int
        Port to connect to.
    tb_limit : int
        Maximum number of stack frames to send in error traceback.
    logger : object
        Python Logger object to log to.
    auto_reconnect : bool
        Whether to automatically reconnect if the connection dies.

    Examples
    --------
    >>> MyClient(DeviceClient):
    ...     def reply_myreq(self, msg):
    ...         print str(msg)
    ...
    >>> c = MyClient('localhost', 10000)
    >>> c.start()
    >>> c.send_request(katcp.Message.request('myreq'))
    >>> # expect reply to be printed here
    >>> # stop the client once we're finished with it
    >>> c.stop()
    >>> c.join()

    Notes
    -----

    The client may block its ioloop if the default blocking tornado DNS
    resolver is used. When an ioloop is shared, it would make sense to configure
    one of the non-blocking resolver classes, see
    http://tornado.readthedocs.org/en/latest/netutil.html

    """

    MAX_MSG_SIZE = 2*1024*1024
    """Maximum message size that can be received in bytes.

    If more than MAX_MSG_SIZE bytes are read from the socket without
    encountering a message terminator (i.e. newline), the connection is closed.

    """
    MAX_WRITE_BUFFER_SIZE = 2*MAX_MSG_SIZE
    """Maximum outstanding bytes to be buffered by the server process.

    If more than MAX_WRITE_BUFFER_SIZE bytes are outstanding, the connection is
    closed. Note that the OS also buffers socket writes, so more than
    MAX_WRITE_BUFFER_SIZE bytes may be untransmitted in total.

    """

    MAX_LOOP_LATENCY = 0.03
    """Do not spend more than this many seconds reading pipelined socket data

    IOStream inline-reading can result in ioloop starvation (see
    https://groups.google.com/forum/#!topic/python-tornado/yJrDAwDR_kA).

    """

    def __init__(self, host, port, tb_limit=20, logger=log,
                 auto_reconnect=True):
        self._parser = MessageParser()
        if not host:
            host = "0.0.0.0"
        self._bindaddr = (host, port)
        self._tb_limit = tb_limit
        self._logger = logger
        self._auto_reconnect = auto_reconnect
        # Number of seconds to wait before retrying a connection
        self.auto_reconnect_delay = 0.5
        self._connect_failures = 0
        self._static_protocol_configuration = False
        self._static_protocol_flags = None
        # Last used unique message id counter
        self._last_msg_id = 0
        # The Tornado IOloop to use, set by self.set_ioloop()
        self._ioloop = None
        # ID of Thread that hosts the IOLoop.
        # Used to check that we are running in the ioloop.
        self.ioloop_thread_id = None
        self._ioloop_manager = IOLoopManager(managed_default=True)
        # Indicate thread safety. Managed by self.enable_thread_safety()
        self._threadsafe = False
        # Last time a connection was made in seconds since Unix Epoch
        self.last_connect_time = None
        self._init_state_per_run(ioloop=None)

    def _init_state_per_run(self, ioloop):
        """Initialise all state that can change per start..stop run.

        The client can be started and stopped multiple times.  This method
        initialises all the variables that keep client state related to a
        single run.

        This method must only be called from `__init__()`, and `start()` as
        it modifies the overall state of the client.

        """
        # Indicate that client's main coroutine is not running
        self._stopped = AsyncEvent(ioloop=ioloop)
        self._stopped.set()
        # Indicate that client is running and ready connect
        self._running = AsyncEvent(ioloop=ioloop)
        # Indicate that the client is waiting for a reconnect. Purely for testing.
        self._waiting_to_retry = AsyncEvent(ioloop=ioloop)
        # Indicate that client is connected to its server
        self._connected = AsyncEvent(ioloop=ioloop)
        # Indicate that the client is disconnected
        self._disconnected = AsyncEvent(ioloop=ioloop)
        self._disconnected.set()

        # Indicate that client has received KATCP protocol info from its server
        self._received_protocol_info = AsyncEvent(ioloop=ioloop)
        self._server_supports_ids = False
        self._protocol_flags = None
        if self._static_protocol_configuration:
            self.preset_protocol_flags(self._static_protocol_flags)

        # Current iostream instance, set by _connect()
        self._stream = None
        # tornado.tcpclient.TCPClient TCP connection factory, set by _install()
        self._tcp_client = None
        # Version information as received from the server
        self.versions = ObjectDict()

    @property
    def protocol_flags(self):
        return self._protocol_flags

    @protocol_flags.setter
    def protocol_flags(self, val):
        self._protocol_flags = val
        self._server_supports_ids = self.protocol_flags.supports(
            ProtocolFlags.MESSAGE_IDS)
        self._received_protocol_info.set()

    @property
    def bind_address(self):
        """(host, port) where the client is connecting"""
        return self._bindaddr

    @property
    def bind_address_string(self):
        return address_to_string(self._bindaddr)

    @property
    def sockname(self):
        if self._stream:
            return self._stream.socket.getsockname()
        else:
            return (self._bindaddr[0], -1)

    @property
    def sockname_string(self):
        return address_to_string(self.sockname)

    @property
    def threadsafe(self):
        return self._threadsafe

    @property
    def ioloop(self):
        return self._ioloop

    def convert_seconds(self, time_seconds):
        """Convert a time in seconds to the device timestamp units.

        KATCP v4 and earlier, specified all timestamps in milliseconds. Since
        KATCP v5, all timestamps are in seconds. If the device KATCP version
        has been detected, this method converts a value in seconds to the
        appropriate (seconds or milliseconds) quantity. For version smaller
        than V4, the time value will be truncated to the nearest millisecond.

        """
        if self.protocol_flags.major >= SEC_TS_KATCP_MAJOR:
            return time_seconds
        else:
            device_time = time_seconds * SEC_TO_MS_FAC
            if self.protocol_flags.major < FLOAT_TS_KATCP_MAJOR:
                device_time = int(device_time)
            return device_time

    def _next_id(self):
        """Return the next available message id as a byte string."""
        assert get_thread_ident() == self.ioloop_thread_id
        self._last_msg_id += 1
        return b"%d" % self._last_msg_id

    def preset_protocol_flags(self, protocol_flags):
        """Preset server protocol flags.

        Sets the assumed server protocol flags and disables automatic server
        version detection.

        Parameters
        ----------
        protocol_flags : katcp.core.ProtocolFlags instance

        """
        self._static_protocol_configuration = True
        self._static_protocol_flags = protocol_flags
        self.protocol_flags = protocol_flags

    def inform_version_connect(self, msg):
        """Process a #version-connect message."""
        if len(msg.arguments) < 2:
            return
        # Store version information.
        name = ensure_native_str(msg.arguments[0])
        self.versions[name] = tuple(
            ensure_native_str(arg) for arg in msg.arguments[1:])
        if msg.arguments[0] == b"katcp-protocol":
            protocol_flags = ProtocolFlags.parse_version(msg.arguments[1])
            self._set_protocol_from_inform(protocol_flags, msg)

    def inform_version(self, msg):
        """Handle katcp v4 and below version inform."""
        self._set_v4_protocol(msg)

    def inform_build_state(self, msg):
        """Handle katcp v4 and below build-state inform."""
        self._set_v4_protocol(msg)

    def _set_v4_protocol(self, inform):
        # We don't know if the server supports multiple connections (katcp v4
        # has no way of indicating this), but this should not really make any
        # difference to the client
        protocol_flags = ProtocolFlags(4, 0, '')
        self._set_protocol_from_inform(protocol_flags, inform)

    def _set_protocol_from_inform(self, protocol_flags, inform):
        if protocol_flags == self.protocol_flags:
            # New value matches old, no need to do consistency checking
            return
        if self.protocol_flags:
            # It seems that the protocol flags have been set before.
            # Now we need to do some consistency checking.
            if self._static_protocol_configuration:
                # Only warn if a static protocol definition is used
                self._logger.warn(
                    'Protocol Version Warning: Ignoring inform received from '
                    'server indicating a katcp protocol revision inconsistent '
                    'with the static configuration. Static configuration: %r.'
                    'Inform received: %r' % (
                        str(self.protocol_flags), str(inform)))
                return
            else:
                # Log an error and disconnect if we are in auto-detection mode
                self._logger.error(
                    'Protocol Version Error: Inform received from '
                    'server indicating a katcp protocol revision inconsistent '
                    'with the previously detected version. Disconnecting in '
                    'disgust. Previous version: %r. Inform received: %r' % (
                        str(self.protocol_flags), str(inform)))
                # Prevent an infinite loop
                self._auto_reconnect = False
                self._disconnect()

        self.protocol_flags = protocol_flags

    def _get_mid_and_update_msg(self, msg, use_mid):
        """Get message ID for current request and assign to msg.mid if needed.

        Parameters
        ----------
        msg : katcp.Message ?request message
        use_mid : bool or None

        If msg.mid is None, a new message ID will be created. msg.mid will be
        filled with this ID if use_mid is True or if use_mid is None and the
        server supports message ids. If msg.mid is already assigned, it will
        not be touched, and will be used as the active message ID.

        Return value
        ------------
        The active message ID

        """
        if use_mid is None:
            use_mid = self._server_supports_ids

        if msg.mid is None:
            mid = self._next_id()
            if use_mid:
                msg.mid = mid
            # An internal mid may be needed for the request/inform/response
            # machinery to work, so we return it
            return mid
        else:
            return msg.mid

    @make_threadsafe_blocking
    def request(self, msg, use_mid=None):
        """Send a request message, with automatic message ID assignment.

        Parameters
        ----------
        msg : katcp.Message request message
        use_mid : bool or None, default=None

        Returns
        -------
        mid : string or None
            The message id, or None if no msg id is used

        If use_mid is None and the server supports msg ids, or if use_mid is
        True a message ID will automatically be assigned msg.mid is None.

        if msg.mid has a value, and the server supports msg ids, that value
        will be used. If the server does not support msg ids, KatcpVersionError
        will be raised.

        """
        mid = self._get_mid_and_update_msg(msg, use_mid)
        self.send_request(msg)
        return mid

    def send_request(self, msg):
        """Send a request message.

        Parameters
        ----------
        msg : Message object
            The request Message to send.

        """
        assert(msg.mtype == Message.REQUEST)
        if msg.mid and not self._server_supports_ids:
            raise KatcpVersionError('Message IDs not supported by server')
        self.send_message(msg)

    @make_threadsafe_blocking
    def send_message(self, msg):
        """Send any kind of message.

        Parameters
        ----------
        msg : Message object
            The message to send.

        """
        assert get_thread_ident() == self.ioloop_thread_id
        data = bytes(msg) + b"\n"

        # Log all sent messages here so no one else has to.
        if self._logger.isEnabledFor(logging.DEBUG):
            self._logger.debug("Sending to {}: {}"
                               .format(self.bind_address_string, repr(data)))
        if not self._connected.isSet():
            raise KatcpClientDisconnected('Not connected to device {0}'.format(
                self.bind_address_string))
        try:
            return self._stream.write(data)
        except Exception:
            self._logger.warn('Could not send message {0!r} to {1!r}'
                              .format(str(msg), self._bindaddr), exc_info=True)
            self._disconnect(exc_info=True)

    @gen.coroutine
    def _connect(self):
        """Connect to the server."""
        assert get_thread_ident() == self.ioloop_thread_id
        if self._stream:
            self._logger.warn('Disconnecting existing connection to {0!r} '
                              'to create a new connection')
            self._disconnect()
            yield self._disconnected.until_set()

        stream = None
        try:
            host, port = self._bindaddr
            stream = self._stream = yield self._tcp_client.connect(
                host, port, max_buffer_size=self.MAX_MSG_SIZE)
            stream.set_close_callback(partial(self._stream_closed_callback,
                                              stream))
            # our message packets are small, don't delay sending them.
            stream.set_nodelay(True)
            stream.max_write_buffer_size = self.MAX_WRITE_BUFFER_SIZE
            self._logger.debug('Connected to {0} with client addr {1}'
                               .format(self.bind_address_string,
                                       address_to_string(stream.socket.getsockname())))
            if self._connect_failures >= 5:
                self._logger.warn("Reconnected to {0}"
                                  .format(self.bind_address_string))
            self._connect_failures = 0
        except Exception as e:
            if self._connect_failures % 5 == 0:
                # warn on every fifth failure

                # TODO (NM 2015-03-04) This can get a bit verbose, and typically we have
                # other mechanisms for tracking failed connections. Consider doing some
                # kind of exponential backoff starting at 5 times the reconnect time up to
                # once per 5 minutes
                self._logger.debug("Failed to connect to {0!r}: {1}"
                                  .format(self._bindaddr, e))
            self._connect_failures += 1
            stream = None
            yield gen.moment
            # TODO some kind of error rate limiting?

            if self._stream:
                # Can't use _disconnect() and wait on self._disconnected, since
                # exception may have been raised before _stream_closed_callback
                # was attached to the iostream.
                self._logger.debug('stream was set even though connecting failed')
                self._stream.close()
                self._disconnected.set()

        if stream:
            self._disconnected.clear()
            self._connected.set()
            self.last_connect_time = self.ioloop.time()
            try:
                self.notify_connected(True)
            except Exception:
                self._logger.exception("Notify connect failed. Disconnecting.")
                self._disconnect()

    @make_threadsafe
    def disconnect(self):
        """Force client connection to close, reconnect if auto-connect set."""
        self._disconnect()

    def _disconnect(self, exc_info=False):
        """Disconnect and cleanup."""
        if self._stream:
            self._stream.close(exc_info=exc_info)
            # Cleanup should be handled by _stream_closed_callback()

    def _stream_closed_callback(self, stream):
        try:
            assert stream is self._stream
            self._stream = None
            self._connected.clear()
            self._disconnected.set()
            error_repr = '{0!r}'.format(stream.error) if stream.error else ''
            if error_repr:
                self._logger.warn('Client stream for server '
                                  '{0} closed with error {1}'
                                  .format(self.bind_address_string, error_repr))
            try:
                self.notify_connected(False)
            except Exception:
                self._logger.exception("Notify connect failed")
        except Exception:
            self._logger.exception('Unhandled exception closing client '
                                   'stream {0!r} to {1}'
                                   .format(stream, self.bind_address_string))

    def handle_message(self, msg):
        """Handle a message from the server.

        Parameters
        ----------
        msg : Message object
            The Message to dispatch to the handler methods.

        """
        # log messages received so that no one else has to
        if self._logger.isEnabledFor(logging.DEBUG):
            self._logger.debug(
                "received from {}: {}"
                .format(self.bind_address_string, repr(str(msg))))

        if msg.mtype == Message.INFORM:
            return self.handle_inform(msg)
        elif msg.mtype == Message.REPLY:
            return self.handle_reply(msg)
        elif msg.mtype == Message.REQUEST:
            return self.handle_request(msg)
        else:
            self._logger.error("Unexpected message type from server ['%s']."
                               % (msg,))

    def handle_inform(self, msg):
        """Dispatch an inform message to the appropriate method.

        Parameters
        ----------
        msg : Message object
            The inform message to dispatch.

        """
        method = self._inform_handlers.get(
            msg.name, self.__class__.unhandled_inform)

        try:
            return method(self, msg)
        except Exception:
            e_type, e_value, trace = sys.exc_info()
            reason = "\n".join(traceback.format_exception(
                e_type, e_value, trace, self._tb_limit))
            self._logger.error("Inform %s FAIL: %s" % (msg.name, reason))

    def handle_reply(self, msg):
        """Dispatch a reply message to the appropriate method.

        Parameters
        ----------
        msg : Message object
            The reply message to dispatch.

        """
        method = self.__class__.unhandled_reply
        if msg.name in self._reply_handlers:
            method = self._reply_handlers[msg.name]

        try:
            return method(self, msg)
        except Exception:
            e_type, e_value, trace = sys.exc_info()
            reason = "\n".join(traceback.format_exception(
                e_type, e_value, trace, self._tb_limit))
            self._logger.error("Reply %s FAIL: %s" % (msg.name, reason))

    def handle_request(self, msg):
        """Dispatch a request message to the appropriate method.

        Parameters
        ----------
        msg : Message object
            The request message to dispatch.

        """
        method = self.__class__.unhandled_request
        if msg.name in self._request_handlers:
            method = self._request_handlers[msg.name]

        try:
            reply = method(self, msg)
            if isinstance(reply, Message):
                # If it is a message object, assume it is a reply.
                reply.mid = msg.mid
                assert (reply.mtype == Message.REPLY)
                assert (reply.name == msg.name)
                self._logger.info("%s OK" % (msg.name,))
                self.send_message(reply)
            else:
                # Just pass on what is, potentially, a future. The implementer
                # of the request handler method must arrange for a reply to be
                # sent. Since clients have no business dealing with requests in
                # any case we don't do much to help them.
                return reply
        # We do want to catch everything that inherits from Exception
        # pylint: disable-msg = W0703
        except Exception:
            e_type, e_value, trace = sys.exc_info()
            reason = "\n".join(traceback.format_exception(
                e_type, e_value, trace, self._tb_limit))
            self._logger.error("Request %s FAIL: %s" % (msg.name, reason))

    def unhandled_inform(self, msg):
        """Fallback method for inform messages without a registered handler.

        Parameters
        ----------
        msg : Message object
            The inform message that wasn't processed by any handlers.

        """
        pass

    def unhandled_reply(self, msg):
        """Fallback method for reply messages without a registered handler.

        Parameters
        ----------
        msg : Message object
            The reply message that wasn't processed by any handlers.

        """
        pass

    def unhandled_request(self, msg):
        """Fallback method for requests without a registered handler.

        Parameters
        ----------
        msg : Message object
            The request message that wasn't processed by any handlers.

        """
        pass

    @gen.coroutine
    def _install(self):
        self._stopped.clear()
        try:
            # Do stuff to put us on the IOLoop
            self._logger.debug("Starting client loop for {0!r}"
                               .format(self._bindaddr))
            self.ioloop_thread_id = get_thread_ident()
            self._tcp_client = tornado.tcpclient.TCPClient()
            self._running.set()
            yield self._connect()
            if self._stream is None and not self._auto_reconnect:
                raise KatcpClientError("Failed to connect to {0!r}"
                                       .format(self._bindaddr))
            try:
                yield self._connect_loop()
            except Exception:
                self._logger.exception('Unhandled exception in _connect_loop()')
            finally:
                try:
                    # Make sure everything is torn down properly
                    if self.running():
                        self.stop()
                except RuntimeError as e:
                    if str(e) == 'IOLoop is closing':
                        # Seems the ioloop was stopped already, no worries.
                        self._running.clear()
                self._logger.info("Client connect loop for {0!r} finished."
                                  .format(self._bindaddr))
        finally:
            self._stopped.set()

    @gen.coroutine
    def _connect_loop(self):
        while self._running.isSet():
            if self._connected.isSet():
                try:
                    yield self._line_read_loop()
                except Exception:
                    self._logger.exception('Unhandled exception in _reading_loop()')
            elif self._auto_reconnect:
                self.ioloop.add_callback(self._waiting_to_retry.set)
                yield until_later(self.auto_reconnect_delay)
                self._waiting_to_retry.clear()
                # client may have been stopped while waiting on reconnect delay
                if self.running():
                    yield self._connect()
            else:
                break
            # Allow stream-close handlers etc to run
            # before restarting _line_read_loop
            yield gen.moment

    @gen.coroutine
    def _line_read_loop(self):
        assert get_thread_ident() == self.ioloop_thread_id
        latency_timer = LatencyTimer(self.MAX_LOOP_LATENCY)
        while self._running.isSet():
            try:
                line_fut = self._stream.read_until_regex(b'\n|\r')
                latency_timer.check_future(line_fut)
                if latency_timer.time_to_yield():
                    yield gen.moment
                line = yield line_fut
            except tornado.iostream.StreamClosedError:
                # Assume that _stream_closed_callback() will handle this case
                break
            except Exception:
                if self._stream:
                    line = b''
                    self._logger.warn('Unhandled Exception while reading from {0}:'
                                      .format(self._bindaddr), exc_info=True)
                    # Prevent potential tight error loops from blocking ioloop
                    self._disconnect()
                    yield gen.moment
                else:
                    self._logger.warn('self._stream object seems to have disappeared.')
                    break
            try:
                line = line.replace(b"\r", b"\n").split(b"\n")[0]
                msg = self._parser.parse(line) if line else None
            except Exception:
                e_type, e_value, trace = sys.exc_info()
                reason = "\n".join(traceback.format_exception(
                    e_type, e_value, trace, self._tb_limit))
                self._logger.error("BAD COMMAND: %s" % (reason,))
            else:
                try:
                    if msg:
                        yield gen.maybe_future(self.handle_message(msg))
                except Exception:
                    self._logger.exception(
                        'Unhandled exception in handle_message() from {0} for '
                        'message {1!r}'.format(self.bind_address_string, str(msg)))

        self._disconnect()

        self._logger.debug('client _line_read_loop() from {0} completed'
                           .format(self.bind_address_string))

    def set_ioloop(self, ioloop=None):
        """Set the tornado.ioloop.IOLoop instance to use.

        This defaults to IOLoop.current(). If set_ioloop() is never called the
        IOLoop is managed: started in a new thread, and will be stopped if
        self.stop() is called.

        Notes
        -----
        Must be called before start() is called

        """
        self._ioloop_manager.set_ioloop(ioloop, managed=False)
        self._ioloop = ioloop

    def enable_thread_safety(self):
        """Enable thread-safety features.

        Must be called before start().

        """
        if self.threadsafe:
            return                        # Already done!
        if self._running.isSet():
            raise RuntimeError('Cannot enable thread safety after start')

        def _getattr(obj, name):
            # use 'is True' so mock objects don't return true for everything
            return getattr(obj, name, False) is True

        for name in dir(self):
            try:
                meth = getattr(self, name)
            except AttributeError:
                # Subclasses may have computed attributes that don't work
                # before they are started, so let's ignore those
                pass
            if not callable(meth):
                continue
            make_threadsafe = _getattr(meth, 'make_threadsafe')
            make_threadsafe_blocking = _getattr(meth, 'make_threadsafe_blocking')

            if make_threadsafe:
                assert not make_threadsafe_blocking
                meth = self._make_threadsafe(meth)
                setattr(self, name, meth)
            elif make_threadsafe_blocking:
                meth = self._make_threadsafe_blocking(meth)
                setattr(self, name, meth)

        self._threadsafe = True

    def _make_threadsafe(self, meth):
        @wraps(meth)
        def wrapped(*args, **kwargs):
            if get_thread_ident() == self.ioloop_thread_id:
                return meth(*args, **kwargs)
            else:
                self.ioloop.add_callback(meth, *args, **kwargs)
        return wrapped

    def _make_threadsafe_blocking(self, meth):
        @wraps(meth)
        def wrapped(*args, **kwargs):
            timeout = kwargs.pop('timeout', 5)
            if get_thread_ident() == self.ioloop_thread_id:
                return meth(*args, **kwargs)
            else:
                f = Future()
                def cb():
                    try:
                        tf = gen.maybe_future(meth(*args, **kwargs))
                    except Exception:
                        tf = tornado_Future()
                        tf.set_exc_info(sys.exc_info())
                    finally:
                        gen.chain_future(tf, f)

                self.ioloop.add_callback(cb)
                return f.result(timeout)
        return wrapped

    def start(self, timeout=None):
        """Start the client in a new thread.

        Parameters
        ----------
        timeout : float in seconds
            Seconds to wait for client thread to start. Do not specify a
            timeout if start() is being called from the same ioloop that this
            client will be installed on, since it will block the ioloop without
            progressing.

        """
        if self._running.isSet():
            raise RuntimeError("Device client already started.")
        # Make sure we have an ioloop
        self._ioloop = self._ioloop_manager.get_ioloop()
        if timeout:
            t0 = self.ioloop.time()
        self._ioloop_manager.start(timeout)
        # update state variables before starting `_install()`
        self._init_state_per_run(ioloop=self.ioloop)
        self.ioloop.add_callback(self._install)
        if timeout:
            remaining_timeout = timeout - (self.ioloop.time() - t0)
            self.wait_running(remaining_timeout)

    def join(self, timeout=None):
        """Rejoin the client thread.

        Parameters
        ----------
        timeout : float in seconds
            Seconds to wait for thread to finish.

        Notes
        -----
        Does nothing if the ioloop is not managed.  Use until_stopped() instead.

        """
        self._ioloop_manager.join(timeout)

    def stop(self, timeout=None):
        """Stop a running client.

        If using a managed ioloop, this must be called from a different
        thread to the ioloop's.  This method only returns once the client's
        main coroutine, `_install()`, has completed.

        If using an unmanaged ioloop, this can be called from the
        same thread as the ioloop.  The `until_stopped()` method can be
        used to wait on completion of the main coroutine, `_install()`.

        Parameters
        ----------
        timeout : float in seconds
           Seconds to wait for both client thread to have *started*, and
           for stopping.

        """
        ioloop = getattr(self, 'ioloop', None)
        if not ioloop:
            raise RuntimeError('Call start() before stop()')
        stopped_from_ioloop_thread = get_thread_ident() == self.ioloop_thread_id

        @tornado.gen.coroutine
        def _stop():
            if timeout:
                yield self._running.until_set(timeout)
            # clear _running before disconnecting to allow line_read_loop
            # and connect_loop to exit
            self._running.clear()
            self._disconnect()
            if not stopped_from_ioloop_thread:
                yield self._stopped.until_set(timeout)
            # else unmanaged ioloop, so caller should yield on until_stopped()

        self._ioloop_manager.stop(timeout=timeout, callback=_stop)

    def until_stopped(self, timeout=None):
        """Return future that resolves when the client has stopped.

        Parameters
        ----------
        timeout : float in seconds
            Seconds to wait for the client to stop.

        Notes
        -----

        If already running, `stop()` must be called before this.

        Must be called from the same ioloop as the client.
        If using a different thread, or a managed ioloop, this
        method should not be used.  Use `join()` instead.

        Also note that stopped != not running.  Stopped means the
        main coroutine has ended, or was never started.  When stopping,
        the running flag is cleared some time before stopped is set.

        """
        assert get_thread_ident() == self.ioloop_thread_id
        return self._stopped.until_set(timeout=timeout)

    def running(self):
        """Whether the client is running.

        Returns
        -------
        running : bool
            Whether the client is running.

        """
        return self._running.isSet()

    def until_running(self, timeout=None):
        """Return future that resolves when the client is running.

        Notes
        -----

        Must be called from the same ioloop as the client.

        """
        return self._running.until_set(timeout=timeout)

    def wait_running(self, timeout=None):
        """Wait until the client is running.

        Parameters
        ----------
        timeout : float in seconds
            Seconds to wait for the client to start running.

        Returns
        -------
        running : bool
            Whether the client is running

        Notes
        -----
        Do not call this from the ioloop, use until_running().

        """
        ioloop = getattr(self, 'ioloop', None)
        if not ioloop:
            raise RuntimeError('Call start() before wait_running()')
        return self._running.wait_with_ioloop(ioloop, timeout)

    def is_connected(self):
        """Check if the socket is currently connected.

        Returns
        -------
        connected : bool
            Whether the client is connected.

        """
        return self._connected.isSet()

    @tornado.gen.coroutine
    def until_connected(self, timeout=None):
        """Return future that resolves when the client is connected."""
        t0 = self.ioloop.time()
        yield self.until_running(timeout=timeout)
        t1 = self.ioloop.time()
        if timeout:
            timedelta = timeout - (t1 - t0)
        else:
            timedelta = None
        assert get_thread_ident() == self.ioloop_thread_id
        yield self._connected.until_set(timeout=timedelta)

    def wait_connected(self, timeout=None):
        """Wait until the client is connected.

        Parameters
        ----------
        timeout : float in seconds
            Seconds to wait for the client to connect.

        Returns
        -------
        connected : bool
            Whether the client is connected.

        Notes
        -----
        Do not call this from the ioloop, use until_connected().

        """
        return self._connected.wait_with_ioloop(self.ioloop, timeout)

    def wait_disconnected(self, timeout=None):
        """Wait until the client is disconnected.

        Parameters
        ----------
        timeout : float in seconds
            Seconds to wait for the client to disconnect.

        Returns
        -------
        disconnected : bool
            Whether the client is disconnected.

        Notes
        -----
        Do not call this from the ioloop, use until_disconnected().

        """
        return self._disconnected.wait_with_ioloop(self.ioloop, timeout)

    @tornado.gen.coroutine
    def until_protocol(self, timeout=None):
        """Return future that resolves after receipt of katcp protocol info.

        If the returned future resolves, the server's protocol information is
        available in the ProtocolFlags instance self.protocol_flags.

        """
        t0 = self.ioloop.time()
        yield self.until_running(timeout=timeout)
        t1 = self.ioloop.time()
        if timeout:
            timedelta = timeout - (t1 - t0)
        else:
            timedelta = None
        assert get_thread_ident() == self.ioloop_thread_id
        yield self._received_protocol_info.until_set(timeout=timedelta)

    def wait_protocol(self, timeout=None):
        """Wait until katcp protocol information has been received from server.

        Parameters
        ----------
        timeout : float in seconds
            Seconds to wait for the client to connect.

        Returns
        -------
        received : bool
            Whether protocol information was received

        If this method returns True, the server's protocol information is
        available in the ProtocolFlags instance self.protocol_flags.

        Notes
        -----
        Do not call this from the ioloop, use until_protocol().

        """
        return self._received_protocol_info.wait_with_ioloop(self.ioloop, timeout)

    def notify_connected(self, connected):
        """Event handler that is called whenever the connection status changes.

        Override in derived class for desired behaviour.

        .. note::

           This function should never block. Doing so will cause the client to
           cease processing data from the server until notify_connected
           completes.

        Parameters
        ----------
        connected : bool
            Whether the client has just connected (True) or just disconnected
            (False).

        """
        pass


class AsyncClient(DeviceClient):
    """Implement async and callback-based requests on top of DeviceClient.

    This client will use message IDs if the server supports them.

    Parameters
    ----------
    host : string
        Host to connect to.
    port : int
        Port to connect to.
    tb_limit : int, optional
        Maximum number of stack frames to send in error traceback.
    logger : object, optional
        Python Logger object to log to. Default is a logger named 'katcp'.
    auto_reconnect : bool, optional
        Whether to automatically reconnect if the connection dies.
    timeout : float in seconds, optional
        Default number of seconds to wait before a callback callback_request
        times out. Can be overridden in individual calls to callback_request.

    Examples
    --------
    >>> def reply_cb(msg):
    ...     print "Reply:", msg
    ...
    >>> def inform_cb(msg):
    ...     print "Inform:", msg
    ...
    >>> c = AsyncClient('localhost', 10000)
    >>> c.start()
    >>> c.ioloop.add_callback(
    ...     c.callback_request,
    ...     katcp.Message.request('myreq'),
    ...     reply_cb=reply_cb,
    ...     inform_cb=inform_cb,
    ... )
    ...
    >>> # expect reply to be printed here
    >>> # stop the client once we're finished with it
    >>> c.stop()
    >>> c.join()
    """

    def __init__(self, host, port, tb_limit=20, timeout=5.0, logger=log,
                 auto_reconnect=True):
        super(AsyncClient, self).__init__(host, port, tb_limit=tb_limit,
                                          logger=logger,
                                          auto_reconnect=auto_reconnect)

        self._request_timeout = timeout
        self._reset_async_requests()

    def _reset_async_requests(self):
        """Initialize / clear out async request structures.

        Good for a clean start after disconnection.

        """
        # pending requests
        # msg_id -> (request, reply_cb, inform_cb, user_data, timeout_handle)
        #           callback tuples
        self._async_queue = {}

        # stack mapping request names to a stack of message ids
        # msg_name -> [ list of msg_ids ]
        self._async_id_stack = {}

    def _push_async_request(self, msg_id, request, reply_cb, inform_cb,
                            user_data, timeout_handle):
        """Store reply / inform callbacks for request we've sent."""
        assert get_thread_ident() == self.ioloop_thread_id
        self._async_queue[msg_id] = (
            request, reply_cb, inform_cb, user_data, timeout_handle)
        if request.name in self._async_id_stack:
            self._async_id_stack[request.name].append(msg_id)
        else:
            self._async_id_stack[request.name] = [msg_id]

    def _pop_async_request(self, msg_id, msg_name):
        """Pop the set of callbacks for a request.

        Return tuple of Nones if callbacks already popped (or don't exist).

        """
        assert get_thread_ident() == self.ioloop_thread_id
        if msg_id is None:
            msg_id = self._msg_id_for_name(msg_name)
        if msg_id in self._async_queue:
            callback_tuple = self._async_queue[msg_id]
            del self._async_queue[msg_id]
            self._async_id_stack[callback_tuple[0].name].remove(msg_id)
            return callback_tuple
        else:
            return None, None, None, None, None

    def _peek_async_request(self, msg_id, msg_name):
        """Peek at the set of callbacks for a request.

        Return tuple of Nones if callbacks don't exist.

        """
        assert get_thread_ident() == self.ioloop_thread_id
        if msg_id is None:
            msg_id = self._msg_id_for_name(msg_name)
        if msg_id in self._async_queue:
            return self._async_queue[msg_id]
        else:
            return None, None, None, None, None

    def _msg_id_for_name(self, msg_name):
        """Find the msg_id for a given request name.

        Return None if no message id exists.

        """
        if msg_name in self._async_id_stack and self._async_id_stack[msg_name]:
            return self._async_id_stack[msg_name][0]

    @make_threadsafe
    def callback_request(self, msg, reply_cb=None, inform_cb=None,
                         user_data=None, timeout=None, use_mid=None):
        """Send a request message.

        Parameters
        ----------
        msg : Message object
            The request message to send.
        reply_cb : function
            The reply callback with signature reply_cb(msg)
            or reply_cb(msg, \*user_data)
        inform_cb : function
            The inform callback with signature inform_cb(msg)
            or inform_cb(msg, \*user_data)
        user_data : tuple
            Optional user data to send to the reply and inform
            callbacks.
        timeout : float in seconds
            How long to wait for a reply. The default is the
            the timeout set when creating the AsyncClient.
        use_mid : boolean, optional
            Whether to use message IDs. Default is to use message IDs
            if the server supports them.

        """
        if timeout is None:
            timeout = self._request_timeout

        mid = self._get_mid_and_update_msg(msg, use_mid)

        if timeout is None:  # deal with 'no timeout', i.e. None
            timeout_handle = None
        else:
            timeout_handle = self.ioloop.call_later(
                timeout, partial(self._handle_timeout, mid, self.ioloop.time()))

        self._push_async_request(
            mid, msg, reply_cb, inform_cb, user_data, timeout_handle)

        try:
            self.send_request(msg)
        except KatcpClientError as e:
            error_reply = Message.request(msg.name, "fail", str(e))
            error_reply.mid = mid
            self.handle_reply(error_reply)

    def future_request(self, msg, timeout=None, use_mid=None):
        """Send a request message, with future replies.

        Parameters
        ----------
        msg : Message object
            The request Message to send.
        timeout : float in seconds
            How long to wait for a reply. The default is the
            the timeout set when creating the AsyncClient.
        use_mid : boolean, optional
            Whether to use message IDs. Default is to use message IDs
            if the server supports them.

        Returns
        -------
        A tornado.concurrent.Future that resolves with:

        reply : Message object
            The reply message received.
        informs : list of Message objects
            A list of the inform messages received.

        """
        if timeout is None:
            timeout = self._request_timeout

        f = tornado_Future()
        informs = []

        def reply_cb(msg):
            f.set_result((msg, informs))

        def inform_cb(msg):
            informs.append(msg)

        try:
            self.callback_request(msg, reply_cb=reply_cb, inform_cb=inform_cb,
                                  timeout=timeout, use_mid=use_mid)
        except Exception:
            f.set_exc_info(sys.exc_info())
        return f

    def blocking_request(self, msg, timeout=None, use_mid=None):
        """Send a request message and wait for its reply.

        Parameters
        ----------
        msg : Message object
            The request Message to send.
        timeout : float in seconds
            How long to wait for a reply. The default is the
            the timeout set when creating the AsyncClient.
        use_mid : boolean, optional
            Whether to use message IDs. Default is to use message IDs
            if the server supports them.

        Returns
        -------
        reply : Message object
            The reply message received.
        informs : list of Message objects
            A list of the inform messages received.

        """
        assert (get_thread_ident() != self.ioloop_thread_id), (
            'Cannot call blocking_request() in ioloop')
        if timeout is None:
            timeout = self._request_timeout

        f = Future()  # for thread safety
        tf = [None]   # Place-holder for tornado Future for exception tracebacks
        def blocking_request_callback():
            try:
                tf[0] = frf = self.future_request(msg, timeout=timeout,
                                                  use_mid=use_mid)
            except Exception:
                tf[0] = frf = tornado_Future()
                frf.set_exc_info(sys.exc_info())
            gen.chain_future(frf, f)

        self.ioloop.add_callback(blocking_request_callback)

        # We wait on the future result that should be set by the reply
        # handler callback. If this does not occur within the
        # timeout it means something unexpected went wrong. We give it
        # an extra second to deal with (unlikely?) slowness in the
        # rest of the code.
        extra_wait = 1
        wait_timeout = timeout
        if wait_timeout is not None:
            wait_timeout = wait_timeout + extra_wait
        try:
            return f.result(timeout=wait_timeout)
        except TimeoutError:
            raise RuntimeError('Unexpected error: Async request handler did '
                               'not call reply handler within timeout period')
        except Exception:
            # Use the tornado future to give us a usable traceback
            tf[0].result()
            assert False                  # Should not get here

    def handle_inform(self, msg):
        """Handle inform messages related to any current requests.

        Inform messages not related to the current request go up
        to the base class method.

        Parameters
        ----------
        msg : Message object
            The inform message to dispatch.

        """
        # this may also result in inform_cb being None if no
        # inform_cb was passed to the request method.
        if msg.mid is not None:
            _request, _reply_cb, inform_cb, user_data, _timeout_handle = \
                self._peek_async_request(msg.mid, None)
        else:
            request, _reply_cb, inform_cb, user_data, _timeout_handle = \
                self._peek_async_request(None, msg.name)
            if request is not None and request.mid is not None:
                # we sent a mid but this inform doesn't have one
                inform_cb, user_data = None, None

        if inform_cb is None:
            inform_cb = super(AsyncClient, self).handle_inform
            # override user_data since handle_inform takes no user_data
            user_data = None

        try:
            if user_data is None:
                inform_cb(msg)
            else:
                inform_cb(msg, *user_data)
        except Exception:
            e_type, e_value, trace = sys.exc_info()
            reason = "\n".join(traceback.format_exception(
                e_type, e_value, trace, self._tb_limit))
            self._logger.error("Callback inform %s FAIL: %s" %
                               (msg.name, reason))

    def _do_fail_callback(
            self, reason, msg, reply_cb, inform_cb, user_data, timeout_handle):
        """Do callback for a failed request."""
        # this may also result in reply_cb being None if no
        # reply_cb was passed to the request method

        if reply_cb is None:
            # this happens if no reply_cb was passed in to the request
            return

        reason_msg = Message.reply(msg.name, "fail", reason, mid=msg.mid)

        try:
            if user_data is None:
                reply_cb(reason_msg)
            else:
                reply_cb(reason_msg, *user_data)
        except Exception:
            e_type, e_value, trace = sys.exc_info()
            exc_reason = "\n".join(traceback.format_exception(
                e_type, e_value, trace, self._tb_limit))
            self._logger.error("Callback reply during failure %s, %s FAIL: %s" %
                               (reason, msg.name, exc_reason))

    def _handle_timeout(self, msg_id, start_time):
        """Handle a timed-out callback request.

        Parameters
        ----------
        msg_id : uuid.UUID for message
            The name of the reply which was expected.

        """
        msg, reply_cb, inform_cb, user_data, timeout_handle = \
            self._pop_async_request(msg_id, None)
        # We may have been racing with the actual reply handler if the reply
        # arrived close to the timeout expiry,
        # which means the self._pop_async_request() call gave us None's.
        # In this case, just bail.
        #
        # NM 2014-09-17 Not sure if this is true after porting to tornado,
        # but I'm too afraid to remove this code :-/
        if timeout_handle is None:
            return

        reason = "Request {0.name} timed out after {1:f} seconds.".format(
            msg, self.ioloop.time() - start_time)
        self._do_fail_callback(
            reason, msg, reply_cb, inform_cb, user_data, timeout_handle)

    def handle_reply(self, msg):
        """Handle a reply message related to the current request.

        Reply messages not related to the current request go up
        to the base class method.

        Parameters
        ----------
        msg : Message object
            The reply message to dispatch.

        """
        # this may also result in reply_cb being None if no
        # reply_cb was passed to the request method
        if msg.mid is not None:
            _request, reply_cb, _inform_cb, user_data, timeout_handle = \
                self._pop_async_request(msg.mid, None)
        else:
            request, _reply_cb, _inform_cb, _user_data, timeout_handle = \
                self._peek_async_request(None, msg.name)
            if request is not None and request.mid is None:
                # we didn't send a mid so this is the request we want
                _request, reply_cb, _inform_cb, user_data, timeout_handle = \
                    self._pop_async_request(None, msg.name)
            else:
                reply_cb, user_data = None, None

        if timeout_handle is not None:
            self.ioloop.remove_timeout(timeout_handle)

        if reply_cb is None:
            reply_cb = super(AsyncClient, self).handle_reply
            # override user_data since handle_reply takes no user_data
            user_data = None

        try:
            if user_data is None:
                reply_cb(msg)
            else:
                reply_cb(msg, *user_data)
        except Exception:
            e_type, e_value, trace = sys.exc_info()
            reason = "\n".join(traceback.format_exception(
                e_type, e_value, trace, self._tb_limit))
            self._logger.error("Callback reply %s FAIL: %s" %
                               (msg.name, reason))

    def stop(self, *args, **kwargs):
        self.ioloop.add_callback(self._fail_waiting_requests,
                                 'Client stopped before reply was received')
        super(AsyncClient, self).stop(*args, **kwargs)

    def _fail_waiting_requests(self, reason):
        # Fail all requests that have not yet received their replies
        for request_data in list(self._async_queue.values()):
            # Last one should be timeout handle
            timeout_handle = request_data[-1]
            if timeout_handle is not None:
                self.ioloop.remove_timeout(timeout_handle)
            # Do add_callback to prevent callback functions from scheduling
            # new requests before we call _reset_async_requests()
            self.ioloop.add_callback(self._do_fail_callback, reason,
                                     *request_data)

        # Clear out the async data structures
        self._reset_async_requests()

    def _stream_closed_callback(self, stream):
        try:
            super(AsyncClient, self)._stream_closed_callback(stream)
            self._fail_waiting_requests('Connection closed before reply was received')
        except Exception:
            self._logger.exception('Unhandled exception closing client '
                                   'stream {0!r} to {1}'
                                   .format(stream, self.bind_address_string))


def request_check(client, exception, *msg_parms, **kwargs):
    """Make blocking request to client and raise exception if reply is not ok.

    Parameters
    ----------
    client : DeviceClient instance
    exception: Exception class to raise
    *msg_parms : Message parameters sent to the Message.request() call
    **kwargs : Keyword arguments
        Forwards kwargs['timeout'] to client.blocking_request().
        Forwards kwargs['mid'] to Message.request().

    Returns
    -------
    reply, informs : as returned by client.blocking_request

    Raises
    ------
    *exception* passed as parameter is raised if reply.reply_ok() is False

    Notes
    -----
    A typical use-case for this function is to use functools.partial() to bind
    a particular client and exception. The resulting function can then be used
    instead of direct client.blocking_request() calls to automate error
    handling.

    """
    timeout = kwargs.get('timeout', None)
    req_msg = Message.request(*msg_parms)
    if timeout is not None:
        reply, informs = client.blocking_request(req_msg, timeout=timeout)
    else:
        reply, informs = client.blocking_request(req_msg)

    if not reply.reply_ok():
        raise exception('Unexpected failure reply "{2}"\n'
                        ' with device at {0}, request \n"{1}"'
                        .format(client.bind_address_string, req_msg, reply))
    return reply, informs


# Compatibility classes

class CallbackClient(AsyncClient):
    def __init__(self, host, port, tb_limit=20, timeout=5.0, logger=log,
                 auto_reconnect=True):
        super(CallbackClient, self).__init__(host, port, tb_limit=tb_limit,
                                             logger=logger, timeout=timeout,
                                             auto_reconnect=auto_reconnect)
        self.enable_thread_safety()

    def setDaemon(self, daemonic):
        """Set daemonic state of the managed ioloop thread to True / False

        Calling this method for a non-managed ioloop has no effect. Must be called before
        start(), or it will also have no effect
        """
        self._ioloop_manager.setDaemon(daemonic)

class BlockingClient(CallbackClient):
    pass



# Copyright 2016 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function

import tornado

from tornado.ioloop import IOLoop

from katcp import resource_client

ioloop = IOLoop.current()

client = resource_client.KATCPClientResource(dict(
    name='demo-client',
    address=('localhost', 5000),
    controlled=True))

@tornado.gen.coroutine
def demo():
    # Wait until the client has finished inspecting the device
    yield client.until_synced()
    help_response = yield client.req.help()
    print("device help:\n ", help_response)
    add_response = yield client.req.add(3, 6)
    print("3 + 6 response:\n", add_response)
    # By not yielding we are not waiting for the response
    pick_response_future = client.req.pick_fruit()
    # Instead we wait for the fruit.result sensor status to change to
    # nominal. Before we can wait on a sensor, a strategy must be set:
    client.sensor.fruit_result.set_strategy('event')
    # If the condition does not occur within the timeout (default 5s), we will
    # get a TimeoutException
    yield client.sensor.fruit_result.wait(
        lambda reading: reading.status == 'nominal')
    fruit = yield client.sensor.fruit_result.get_value()
    print('Fruit picked: ', fruit)
    # And see how the ?pick-fruit request responded by yielding on its future
    pick_response = yield pick_response_future
    print('pick response: \n', pick_response)

    ioloop.stop()

# Note, katcp.resource_client.ThreadSafeKATCPClientResourceWrapper can be used to
# turn the client into a 'blocking' client for use in e.g. ipython. It will turn
# all functions that return tornado futures into blocking calls, and will bounce
# all method calls through the ioloop. In this case the ioloop must be started
# in a separate thread. katcp.ioloop_manager.IOLoopManager can be used to manage
# the ioloop thread.

ioloop.add_callback(client.start)
ioloop.add_callback(demo)
ioloop.start()



# Copyright 2014 National Research Foundation (South African Radio Astronomy Observatory)
# BSD license - see LICENSE for details

from __future__ import absolute_import, division, print_function
from future import standard_library
standard_library.install_aliases()  # noqa: E402

import copy
import gc
import logging
import time
import unittest
import weakref

from builtins import object
from concurrent.futures import TimeoutError

import mock
import tornado

# module under test
from katcp import (Message, Sensor, ioloop_manager,
                   resource, resource_client)
from katcp.core import AsyncEvent, AttrDict, ProtocolFlags
from katcp.testutils import (DeviceTestSensor, DeviceTestServer,
                             TimewarpAsyncTestCase,
                             TimewarpAsyncTestCaseTimeAdvancer,
                             start_thread_with_cleanup)


logger = logging.getLogger(__name__)


class test_transform_future(tornado.testing.AsyncTestCase):
    def test_transform(self):
        orig_f = tornado.concurrent.Future()
        transform = mock.Mock()
        trans_f = resource_client.transform_future(transform, orig_f)
        retval = mock.Mock()
        orig_f.set_result(retval)
        self.assertIs(trans_f.result(), transform.return_value)
        transform.assert_called_once_with(retval)

    @tornado.testing.gen_test
    def test_exception_in_future(self):
        class AnException(Exception): pass
        @tornado.gen.coroutine
        def raiser():
            raise AnException
        orig_f = raiser()
        transform = mock.Mock()
        trans_f = resource_client.transform_future(transform, orig_f)
        with self.assertRaises(AnException):
            trans_f.result()

    def test_exception_in_transform(self):
        orig_f = tornado.concurrent.Future()
        transform = mock.Mock()
        class AnException(Exception): pass
        transform.side_effect = AnException
        trans_f = resource_client.transform_future(transform, orig_f)
        retval = mock.Mock()
        orig_f.set_result(retval)
        transform.assert_called_once_with(retval)
        with self.assertRaises(AnException):
            trans_f.result()


class test_KATCPClientResourceRequest(unittest.TestCase):
    def setUp(self):
        self.mock_client = mock.Mock()
        self.DUT = resource_client.KATCPClientResourceRequest(
            {'name': 'the-request',
             'description': 'The description',
             'timeout_hint': 33.34},
            self.mock_client)

    def test_init(self):
        self.assertEqual(self.DUT.name, 'the-request')
        self.assertEqual(self.DUT.description, 'The description')
        self.assertEqual(self.DUT.timeout_hint, 33.34)
        # Check that we are registered to the correct ABC
        self.assertIsInstance(self.DUT, resource.KATCPRequest)

    def test_request_with_timeout_hint(self):
        reply = self.DUT('parm1', 2)
        self.mock_client.wrapped_request.assert_called_once_with(
            'the-request', 'parm1', 2, timeout=33.34)
        self.assertIs(reply, self.mock_client.wrapped_request.return_value)

    def test_request_no_timeout_hint(self):
        DUT_no_timeout_hint = resource_client.KATCPClientResourceRequest(
            {'name': 'the-other-request',
             'description': 'The other description',
             'timeout_hint': None},
            self.mock_client)
        reply = DUT_no_timeout_hint('aparm', 3)
        self.mock_client.wrapped_request.assert_called_once_with(
            'the-other-request', 'aparm', 3, timeout=None)
        self.assertIs(reply, self.mock_client.wrapped_request.return_value)


class test_KATCPClientResource(tornado.testing.AsyncTestCase):
    def test_init(self):
        resource_spec = dict(
            name='testdev',
            description='resource for testing',
            address=('testhost', 12345),
            controlled=True)
        DUT = resource_client.KATCPClientResource(dict(resource_spec))
        self.assertEqual(DUT.address, resource_spec['address'])
        self.assertEqual(DUT.state, 'disconnected')
        self.assertEqual(DUT.name, resource_spec['name'])
        self.assertEqual(DUT.description, resource_spec['description'])
        self.assertEqual(DUT.parent, None)
        self.assertEqual(DUT.children, {})
        self.assertEqual(DUT.controlled, True)

        # Now try with a parent and no control
        resource_spec['controlled'] = False
        parent = mock.Mock()
        DUT = resource_client.KATCPClientResource(
            dict(resource_spec), parent=parent)
        self.assertEqual(DUT.parent, parent)
        self.assertEqual(DUT.controlled, False)

    @tornado.testing.gen_test
    def test_dummy_requests(self):

        resource_spec_nodummy = dict(
            name='testdev',
            description='resource for testing',
            address=('testhost', 12345),
            controlled=True)
        resource_spec_dummy = dict(resource_spec_nodummy)
        resource_spec_dummy['dummy_unknown_requests'] = True
        requests = ('req-one', 'req_two')
        DUT_nodummy = self.get_DUT_mock_inspecting_client(
            resource_spec_nodummy)
        DUT_dummy = self.get_DUT_mock_inspecting_client(
            resource_spec_dummy)
        yield DUT_dummy._add_requests(requests)
        yield DUT_nodummy._add_requests(requests)
        # Check dummy flag
        self.assertFalse(DUT_nodummy.dummy_unknown_requests)
        self.assertTrue(DUT_dummy.dummy_unknown_requests)
        # First check that actual requests are handled correctly
        for DUT in (DUT_nodummy, DUT_dummy):
            # For real requests we expect a string, see
            # get_DUT_mock_inspecting_client() below.
            req = DUT_nodummy.req.req_one
            self.assertEqual(req.name, 'req-one')

        # Check that the non-dummy client doesn't have non-existing requests
        with self.assertRaises(AttributeError):
            DUT_nodummy.req.blah

        # Check that we get a dummy request for the dummied client
        dummy_req = DUT_dummy.req.blah
        dummy_reply = yield dummy_req('abc', 'def', 123)
        self.assertTrue(dummy_reply.succeeded)

        # Repeat dummy tests for a simple ClientGroup
        group = resource_client.ClientGroup('group', (DUT_nodummy, DUT_dummy))
        # A real request should appear on the group level too
        req = group.req.req_one
        self.assertEqual(req.name, 'req_one')
        # Since group contains at least one dummy client, it too has dummy requests
        dummy_req = group.req.blah
        dummy_reply = yield dummy_req('abc', 'def', 123)
        self.assertTrue(dummy_reply.succeeded)
        # Check that group without dummy clients doesn't have non-existing requests
        group_nodummy = resource_client.ClientGroup('group', (DUT_nodummy,))
        with self.assertRaises(AttributeError):
            group_nodummy.req.blah

    def get_DUT_mock_inspecting_client(self, resource_spec, *args, **kwargs):
        """Return a KATCPClientResource instance with a mocked inspecting client

        Note that the inspecting client request factory is moced to return a
        string matching the name of the request rather than a KATCPRequest object

        """
        DUT = resource_client.KATCPClientResource(
            dict(resource_spec), *args, **kwargs)
        ic = DUT._inspecting_client = mock.Mock()
        def future_get_request(key):
            f = tornado.concurrent.Future()
            req_obj = resource_client.KATCPClientResourceRequest(
                dict(name=key, description=key, timeout_hint=None), ic)
            f.set_result(req_obj)
            return f
        ic.future_get_request.side_effect = future_get_request
        return DUT

    @tornado.testing.gen_test
    def test_control(self):
        always_allow = ('req-one', 'req_two', 'exclude_one')
        always_exclude = ('exclude_one', 'exclude-two')
        normal = ('normal', 'another-normal')
        def katcp_form(reqs):
            return tuple(r.replace('_', '-') for r in reqs)

        dev_requests = set(katcp_form(always_allow + always_exclude + normal))

        resource_spec = dict(
            name='testdev',
            address=('testhost', 12345),
            always_allowed_requests=always_allow,
            always_excluded_requests=always_exclude,
            controlled=True)

        DUT = self.get_DUT_mock_inspecting_client(resource_spec)
        yield DUT._add_requests(dev_requests)
        # We expect all the requests, except for those in the always_exclude list to be
        # available. Note, exclude-one should not be available even though it is in
        # always_allow, since always_exclude overrides always_allow.
        self.assertEqual(sorted(DUT.req),
                         sorted(['req_one', 'req_two', 'normal', 'another_normal']))

        # Now try one with no control, only req-one and req-two should be available
        resource_spec['controlled'] = False
        DUT = self.get_DUT_mock_inspecting_client(resource_spec)
        yield DUT._add_requests(dev_requests)
        self.assertEqual(sorted(DUT.req), sorted(['req_one', 'req_two']))

    @tornado.testing.gen_test
    def test_lowlevel_client_attributes(self):
        resource_spec = dict(
            name='testdev',
            description='resource for testing',
            address=('testhost', 12345),
            controlled=True)
        DUT = resource_client.KATCPClientResource(dict(resource_spec))
        with self.assertRaises(RuntimeError):
            # Before calling start() a runtime error should be raised since the inspecting
            # client has not yet been instantiated
            DUT.versions
        with self.assertRaises(RuntimeError):
            DUT.last_connect_time

        ic = DUT.inspecting_client_factory(DUT.address[0], DUT.address[1], None)
        DUT._inspecting_client = mock.Mock(spec_set=ic)
        DUT._inspecting_client.katcp_client = mock.Mock(spec_set=ic.katcp_client)
        v = DUT._inspecting_client.katcp_client.versions = mock.Mock()
        l = DUT._inspecting_client.katcp_client.last_connect_time = mock.Mock()
        self.assertIs(DUT.versions, v)
        self.assertIs(DUT.last_connect_time, l)

    @tornado.testing.gen_test
    def test_list_sensors(self):

        resource_spec = dict(
            name='testdev',
            address=('testhost', 12345))
        DUT = resource_client.KATCPClientResource(resource_spec)
        sens_manager = mock.create_autospec(
            resource_client.KATCPClientResourceSensorsManager(mock.Mock(), "test"))
        test_sensors_info = AttrDict(
            sens_one=AttrDict(name='sens-one', description='sensor one', value=1),
            sens_two=AttrDict(name='sens.two', description='sensor one', value=2),
            sens_three=AttrDict(name='sens_three', description='sensor three', value=3))
        sensor_strategies = dict(sens_one='event', sens_three='period 10')

        def make_test_sensors(sensors_info):
            test_sensors = AttrDict()
            for sens_pyname, info in sensors_info.items():
                info = dict(info)
                info['sensor_type'] = Sensor.INTEGER
                val = info.pop('value')
                timestamp = val*10
                received_timestamp = timestamp + 1
                sens = test_sensors[sens_pyname] = resource.KATCPSensor(
                    info, sens_manager)
                sens._reading = resource.KATCPSensorReading(
                    received_timestamp, timestamp, Sensor.NOMINAL, val)
                test_sensors[sens_pyname] = sens
            return test_sensors

        test_sensors = make_test_sensors(test_sensors_info)

        sens_manager.get_sampling_strategy.side_effect = (
            lambda sens_name: resource.normalize_strategy_parameters(
                sensor_strategies.get(
                    resource.escape_name(sens_name), 'none')))

        DUT.sensor.update(test_sensors)

        # Simple search based on python identifier
        result = yield DUT.list_sensors('sens_one')
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], resource.SensorResultTuple(
            test_sensors.sens_one, test_sensors_info.sens_one.name,
            'sens_one', test_sensors_info.sens_one.description, 'integer', '',
            test_sensors.sens_one.reading))

        # Now get all the sensors
        result = yield DUT.list_sensors('')
        # built-in `sorted()` and `list.sort()` use __cmp__ for ordering in Python2.
        # However, this breaks compatibility in Python3 due to
        # https://docs.python.org/3/whatsnew/3.0.html#ordering-comparisons
        result.sort(key=lambda obj: obj.name)
        expected_result = sorted([
            resource.SensorResultTuple(
                test_sensors[s_id], test_sensors_info[s_id].name,
                s_id, test_sensors_info[s_id].description, 'integer', '',
                test_sensors[s_id].reading
            )
            for s_id in test_sensors_info
            ], key=lambda obj: obj.name)

        self.assertEqual(result, expected_result)

        # Test that all sensors are found using their Python identifiers
        result = yield DUT.list_sensors('sens_two')
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].object, test_sensors.sens_two)
        result = yield DUT.list_sensors('sens_three')
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].object, test_sensors.sens_three)

        # Test using actual sensor name
        result = yield DUT.list_sensors('sens_one', use_python_identifiers=False)
        self.assertEqual(len(result), 0)
        result = yield DUT.list_sensors('sens-one', use_python_identifiers=False)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].name, 'sens-one')

        # Now test with strategy filter
        result = yield DUT.list_sensors('', strategy=True)
        self.assertEqual(len(result), len(sensor_strategies))

    def test_until_sync_states(self):
        resource_spec = dict(
            name='testdev',
            address=('testhost', 12345))
        DUT = resource_client.KATCPClientResource(resource_spec)

        # We expect the initial state to be 'disconnected', which means until_synced()
        # should return an unresolved future and until_not_synced() a resolved future
        self.assertEqual(DUT.state, 'disconnected')
        self.assertFalse(DUT.until_synced().done())
        self.assertTrue(DUT.until_not_synced().done())

        # Force state to 'syncing', same expectation as for 'disconnected'
        DUT._state.set_state('syncing')
        self.assertFalse(DUT.until_synced().done())
        self.assertTrue(DUT.until_not_synced().done())

        # Force state to 'synced', opposite expectation as for 'disconnected'
        DUT._state.set_state('synced')
        self.assertTrue(DUT.until_synced().done())
        self.assertFalse(DUT.until_not_synced().done())


class test_KATCPClientResource_Integrated(tornado.testing.AsyncTestCase):
    def setUp(self):
        super(test_KATCPClientResource_Integrated, self).setUp()
        self.server = DeviceTestServer('', 0)
        start_thread_with_cleanup(self, self.server)
        self.host, self.port = self.server.bind_address
        self.default_resource_spec = dict(
            name='thething',
            address=self.server.bind_address,
            controlled=True)

    @tornado.gen.coroutine
    def _get_DUT_and_sync(self, resource_spec):
        DUT = resource_client.KATCPClientResource(resource_spec)
        DUT.start()
        yield DUT.until_state('synced')
        raise tornado.gen.Return(DUT)

    @tornado.testing.gen_test(timeout=1)
    def test_requests(self):
        DUT = yield self._get_DUT_and_sync(self.default_resource_spec)
        # Check that all the test-device requests are listed
        self.assertEqual(sorted(DUT.req),
                         sorted(n.replace('-', '_')
                                for n in self.server.request_names))

    @tornado.testing.gen_test(timeout=1)
    def test_active(self):
        DUT = yield self._get_DUT_and_sync(self.default_resource_spec)
        self.assertTrue(DUT.is_active(), 'Expect DUT to be active initialy')
        reply = yield DUT.req.new_command()
        self.assertTrue(reply.succeeded, 'Expect request to be succesful in active state')

        # Set DUT to 'inactive'
        DUT.set_active(False)
        with self.assertRaises(resource.KATCPResourceInactive):
            # Should raise if we attempt to do the request when inactive
            yield DUT.req.new_command()

        # Set DUT to back to 'active'
        DUT.set_active(True)
        reply = yield DUT.req.new_command()
        self.assertTrue(reply.succeeded, 'Expect request to be succesful in active state')


    @tornado.testing.gen_test(timeout=1)
    def test_sensors(self):
       DUT = yield self._get_DUT_and_sync(self.default_resource_spec)
       # Check that all the test-device sensors are listed
       self.assertEqual(sorted(DUT.sensor),
                        sorted(n.replace('-', '_').replace('.', '_')
                               for n in self.server.sensor_names))

    @tornado.testing.gen_test(timeout=1)
    def test_interface_change(self):
        DUT = yield self._get_DUT_and_sync(self.default_resource_spec)
        sensors_before = set(DUT.sensor)
        reqs_before = set(DUT.req)

        # Add a new sensor to the server
        sensor = DeviceTestSensor(DeviceTestSensor.INTEGER, "another.int",
                                  "An Integer.",
                                  "count", [-5, 5], timestamp=self.io_loop.time(),
                                  status=DeviceTestSensor.NOMINAL, value=3)
        self.server.add_sensor(sensor)
        # Check that the sensor does not exist currently
        self.assertNotIn(resource.escape_name(sensor.name), sensors_before)

        # Add a new request to the server
        def request_sparkling_new(self, req, msg):
            """A new command."""
            return Message.reply(msg.name, "ok", "bling1", "bling2")
        self.server._request_handlers['sparkling-new'] = request_sparkling_new
        # Check that the request did not exist before
        self.assertNotIn('sparkling-new', reqs_before)

        # Issue #interface-changed
        self.server.mass_inform(Message.inform('interface-changed'))
        yield DUT.until_state('syncing')
        yield DUT.until_state('synced')

        # Check if sensor/request was added
        self.assertEqual(set(DUT.sensor) - sensors_before, set(['another_int']))
        self.assertEqual(set(DUT.req) - reqs_before, set(['sparkling_new']))

        # And now remove them again
        self.server._request_handlers.pop('sparkling-new')
        self.server.remove_sensor('another.int')

        # Issue #interface-changed
        self.server.mass_inform(Message.inform('interface-changed'))
        yield DUT.until_state('syncing')
        yield DUT.until_state('synced')

        # Check if sensor/request was removed
        self.assertEqual(set(DUT.sensor), sensors_before)
        self.assertEqual(set(DUT.req), reqs_before)

    @tornado.testing.gen_test
    def test_no_memory_leak_after_usage(self):
        DUT = yield self._get_DUT_and_sync(self.default_resource_spec)
        wr = weakref.ref(DUT)

        DUT.stop()
        yield DUT.until_stopped()

        # clear strong reference and check if object can be garbage collected
        DUT = None
        gc.collect()
        self.assertIsNone(wr())


class test_KATCPClientResource_IntegratedTimewarp(TimewarpAsyncTestCase):
    def setUp(self):
        super(test_KATCPClientResource_IntegratedTimewarp, self).setUp()
        self.server = DeviceTestServer('', 0)
        start_thread_with_cleanup(self, self.server)
        self.host, self.port = self.server.bind_address
        self.default_resource_spec = dict(
            name='thething',
            address=self.server.bind_address,
            controlled=True)

    @tornado.gen.coroutine
    def _get_DUT_and_sync(self, resource_spec):
        DUT = resource_client.KATCPClientResource(self.default_resource_spec)
        DUT.start()
        yield DUT.until_state('synced')
        raise tornado.gen.Return(DUT)

    @tornado.testing.gen_test
    def test_disconnect(self):
        # Test that a device disconnect / reconnect is correctly handled
        DUT = yield self._get_DUT_and_sync(self.default_resource_spec)
        initial_reqs = set(DUT.req)
        initial_sensors = set(DUT.sensor)
        self.server.stop()
        self.server.join(timeout=1)
        yield DUT.until_state('disconnected')
        self.assertEqual(DUT.state, 'disconnected')

        # Test that requests fail
        rep = yield DUT.req.watchdog()
        self.assertFalse(rep.succeeded)

        # Restart device so that we can reconnect
        self.server.start()
        # timewarp beyond reconect delay
        self.set_ioloop_time(self.ioloop_time + 1)
        yield DUT.until_state('syncing')
        yield DUT.until_state('synced')
        # check that sensors / requests are unchanged
        self.assertEqual(set(DUT.req), initial_reqs)
        self.assertEqual(set(DUT.sensor), initial_sensors)

        # Now disconnect and change the device, to check that it is properly resynced.
        self.server.stop()
        self.server.join(timeout=1)
        yield DUT.until_state('disconnected')

        # Add a new request to the server
        def request_sparkling_new(self, req, msg):
            """A new command."""
            return Message.reply(msg.name, "ok", "bling1", "bling2")
        self.server._request_handlers['sparkling-new'] = request_sparkling_new
        # Check that the request does not exist currently
        self.assertNotIn('sparkling_new', initial_reqs)

        # Add a new sensor to the server
        sensor = DeviceTestSensor(DeviceTestSensor.INTEGER, "another.int",
                                  "An Integer.",
                                  "count", [-5, 5], timestamp=self.io_loop.time(),
                                  status=DeviceTestSensor.NOMINAL, value=3)
        self.server.add_sensor(sensor)
        # Check that the sensor does not exist currently
        escaped_new_sensor = resource.escape_name(sensor.name)
        self.assertNotIn(resource.escape_name(sensor.name), initial_sensors)

        # Restart device so that we can reconnect
        self.server.start()
        # timewarp beyond reconect delay
        self.set_ioloop_time(self.ioloop_time + 1)
        yield DUT.until_state('syncing')
        yield DUT.until_state('synced')
        # check that sensors / requests are correctly updated
        self.assertEqual(set(DUT.req), initial_reqs | set(['sparkling_new']))
        self.assertEqual(set(DUT.sensor), initial_sensors | set([escaped_new_sensor]))

    @tornado.testing.gen_test(timeout=1000)
    def test_set_sampling_strategy(self):
        self.server.stop()
        self.server.join()
        DUT = resource_client.KATCPClientResource(self.default_resource_spec)
        DUT.start()
        yield tornado.gen.moment
        test_strategy = ('period', '2.5')
        yield DUT.set_sampling_strategy('an_int', test_strategy)
        self.assertEqual(DUT._sensor_strategy_cache['an_int'], ('period', '2.5'))
        # Double-check that the sensor does not yet exist
        self.assertNotIn('an_int', DUT.sensor)
        self.server.start()
        self.server.wait_running(timeout=1)
        advancer = TimewarpAsyncTestCaseTimeAdvancer(self, quantum=0.55)
        advancer.start()
        yield DUT.until_synced()
        self.assertEqual(DUT.sensor.an_int.sampling_strategy, test_strategy)

        # Now call set_sampling_strategy with a different strategy and check that it is
        # applied to the real sensor
        new_test_strategy = ('event',)
        yield DUT.set_sampling_strategy('an_int', new_test_strategy)
        self.assertEqual(DUT.sensor.an_int.sampling_strategy, new_test_strategy)
        self.assertEqual(DUT._sensor_strategy_cache['an_int'], ('event',))

    @tornado.testing.gen_test(timeout=1000)
    def test_set_sampling_strategies(self):
        self.server.stop()
        self.server.join()
        DUT = resource_client.KATCPClientResource(self.default_resource_spec)
        DUT.start()
        yield tornado.gen.moment
        test_strategy = ('period', '2.5')
        yield DUT.set_sampling_strategy('an_int', test_strategy)
        self.assertEqual(DUT._sensor_strategy_cache['an_int'], ('period', '2.5'))
        # Double-check that the sensor does not yet exist
        self.assertNotIn('an_int', DUT.sensor)
        self.server.start()
        self.server.wait_running(timeout=1)
        advancer = TimewarpAsyncTestCaseTimeAdvancer(self, quantum=0.55)
        advancer.start()
        yield DUT.until_synced()
        self.assertEqual(DUT.sensor.an_int.sampling_strategy, test_strategy)

        # Now call set_sampling_strategy with a different strategy and check that it is
        # applied to the real sensor
        new_test_strategy = ('event',)
        yield DUT.set_sampling_strategies('int', new_test_strategy)
        self.assertEqual(DUT.sensor.an_int.sampling_strategy, new_test_strategy)
        self.assertEqual(DUT._sensor_strategy_cache['an_int'], ('event',))

    @tornado.testing.gen_test(timeout=1000)
    def test_drop_sampling_strategy(self):
        self.server.stop()
        self.server.join()
        DUT = resource_client.KATCPClientResource(self.default_resource_spec)
        DUT.start()
        yield tornado.gen.moment

        # Check no error for a non-existant sensor
        DUT.drop_sampling_strategy('not_a_valid_sensor')

        test_strategy = ('period', '2.5')
        yield DUT.set_sampling_strategy('an_int', test_strategy)
        self.assertEqual(DUT._sensor_strategy_cache['an_int'], test_strategy)

        # Now drop the strategy from the cache
        DUT.drop_sampling_strategy('an_int')
        self.assertNotIn('an_int', DUT._sensor_strategy_cache)
        self.assertNotIn('an.int', DUT._sensor_manager._strategy_cache)

        # Strategy was dropped before starting server, so expect NO match
        self.server.start()
        self.server.wait_running(timeout=1)
        advancer = TimewarpAsyncTestCaseTimeAdvancer(self, quantum=0.55)
        advancer.start()
        yield DUT.until_synced()
        self.assertNotEqual(DUT.sensor.an_int.sampling_strategy, test_strategy)

        # set strategy while running - it must be cached in two places
        # the DUT, and the DUT's sensor manager
        test_strategy = ('period', '3.5')
        yield DUT.set_sampling_strategy('an_int', test_strategy)
        self.assertEqual(DUT._sensor_strategy_cache['an_int'], test_strategy)
        self.assertEqual(DUT._sensor_manager._strategy_cache['an.int'], test_strategy)
        self.assertEqual(DUT.sensor.an_int.sampling_strategy, test_strategy)

        # Disconnect the device
        self.server.stop()
        self.server.join(timeout=1)
        yield DUT.until_state('disconnected')

        # Drop the strategy from the cache
        DUT.drop_sampling_strategy('an_int')
        self.assertNotIn('an_int', DUT._sensor_strategy_cache)
        self.assertNotIn('an.int', DUT._sensor_manager._strategy_cache)

        # restart server and check that strategy was NOT applied
        self.server.start()
        yield DUT.until_synced()
        self.assertNotEqual(DUT.sensor.an_int.sampling_strategy, test_strategy)

    @tornado.testing.gen_test(timeout=1000)
    def test_set_sensor_listener(self):
        self.server.stop()
        self.server.join()
        resource_spec = self.default_resource_spec
        DUT = resource_client.KATCPClientResource(resource_spec)
        DUT.start()
        yield tornado.gen.moment
        test_listener1 = lambda *x : None
        test_listener2 = lambda *y : None
        yield DUT.set_sensor_listener('an_int', test_listener1)
        # Double-check that the sensor does not yet exist
        self.assertNotIn('an_int', DUT.sensor)
        self.server.start()
        self.server.wait_running(timeout=1)
        advancer = TimewarpAsyncTestCaseTimeAdvancer(self, quantum=0.55)
        advancer.start()
        yield DUT.until_synced()
        self.assertTrue(DUT.sensor.an_int.is_listener, test_listener1)

        # Now call set_sensor_lister with a different listener and check that it is
        # also subscribed
        yield DUT.set_sensor_listener('an_int', test_listener2)
        self.assertTrue(DUT.sensor.an_int.is_listener, test_listener2)
        self.assertTrue(DUT.sensor.an_int.is_listener, test_listener1)

    # TODO tests
    #
    # * Sensor strategy re-application
    # * Request through request object, also with timeouts
    # * Sensor callbacks (probably in test_resource.py, no need for full integrated test)


class test_KATCPClientResourceContainer(tornado.testing.AsyncTestCase):
    def setUp(self):
        self.default_spec_orig = dict(clients={
            'client1': dict(address=('client1-addr', 1234), controlled=True),
            'client-2': dict(address=('client2-addr', 1235), controlled=True),
            'another-client': dict(address=('another-addr', 1231), controlled=True)},
                                      name='test-container',
                                      description='container for testing')
        # make a copy in case the test or DUT messes up any of the original dicts.
        self.default_spec = copy.deepcopy(self.default_spec_orig)
        super(test_KATCPClientResourceContainer, self).setUp()

    @tornado.testing.gen_test
    def test_groups(self):
        spec = self.default_spec
        spec['groups'] = dict(group1=['client1', 'another-client'],
                              group2=['client1', 'client-2'],
                              group3=['client1', 'client-2', 'another-client'])
        DUT = resource_client.KATCPClientResourceContainer(copy.deepcopy(spec))
        DUT.add_group('group4', ('client-2', 'another-client'))
        self.assertEqual(sorted(DUT.groups), ['group1', 'group2', 'group3', 'group4'])
        spec['groups']['group4'] = ('client-2', 'another-client')

        # TODO PY3 check if we can get rid of the list() call
        for group_name, group in list(DUT.groups.items()):
            # Smoke test that no errors are raised
            group.req
            # Check that the correct clients are in each group
            self.assertEqual(sorted(client.name for client in group.clients),
                             sorted(spec['groups'][group_name]))

        # now some surgery, mocking _inspecting_client and calling _add_requests manually
        def mock_inspecting_client(client):

            make_fake_requests = lambda mock_client: {
                req: resource_client.KATCPClientResourceRequest(
                    {'name':req,
                     'description': 'Description for {}'.format(req),
                     'timeout_hint': None},
                    mock_client)
                for req in ['req-1', 'req-2', 'req-3']}

            def _install_inspecting_client_mocks(mock_client):
                fake_requests = make_fake_requests(mock_client)

                def future_get_request(key):
                    f = tornado.concurrent.Future()
                    f.set_result(fake_requests[key])
                    return f

                def wrapped_request(request_name, *args, **kwargs):
                    f = tornado.concurrent.Future()
                    retval = resource.KATCPReply(Message.reply(request_name, 'ok'), [])
                    f.set_result(retval)
                    return f

                mock_client.future_get_request.side_effect = future_get_request
                mock_client.wrapped_request.side_effect = wrapped_request
                return future_get_request

            client._inspecting_client = mock_inspecting_client = mock.Mock(
                spec_set=resource_client.ReplyWrappedInspectingClientAsync)
            _install_inspecting_client_mocks(mock_inspecting_client)

            return mock_inspecting_client

        m_i_c_1 = mock_inspecting_client(DUT.children.client1)
        m_i_c_2 = mock_inspecting_client(DUT.children.client_2)
        m_i_c_a = mock_inspecting_client(DUT.children.another_client)

        # TODO PY3 check if we can get rid of the list() call
        normalize_reply = lambda reply: {c:r if r is None else str(r.reply)
                                          for c, r in list(reply.items())}

        yield DUT.children.client1._add_requests(['req-1'])
        g1_reply = yield DUT.groups.group1.req.req_1()
        self.assertEqual(normalize_reply(g1_reply),
                         {'client1': '!req-1 ok', 'another-client': None})
        # Should evaluate false since not all the clients replied
        self.assertFalse(g1_reply)

        yield DUT.children.another_client._add_requests(['req-1'])
        g1_reply = yield DUT.groups.group1.req.req_1()
        self.assertEqual(normalize_reply(g1_reply),
                         {'client1': '!req-1 ok', 'another-client': '!req-1 ok'})
        # Should evaluate True since all the clients replied succesfully
        self.assertTrue(g1_reply)

        yield DUT.children.client_2._add_requests(['req-2'])
        # client-2 is in group2 and group3, so req-2 should now show up.
        self.assertIn('req_2', DUT.groups.group2.req)
        self.assertIn('req_2', DUT.groups.group3.req)
        # Check that the requests weren't accidentally added to another group
        self.assertFalse('req_2' in DUT.groups.group1.req)

    def test_init(self):
        m_logger = mock.Mock()
        DUT = resource_client.KATCPClientResourceContainer(
            self.default_spec, logger=m_logger)
        self.assertEqual(DUT.name, 'test-container')
        self.assertEqual(DUT.description, 'container for testing')
        child_specs = self.default_spec_orig['clients']
        self.assertEqual(sorted(DUT.children),
                         sorted(resource.escape_name(n) for n in child_specs))
        # TODO PY3 check if we can get rid of the list() call
        for child_name, child_spec in list(child_specs.items()):
            child = DUT.children[resource.escape_name(child_name)]
            self.assertEqual(child.name, child_name)
            self.assertEqual(child.parent, DUT)
            self.assertEqual(child.address, child_spec['address'])
            self.assertIs(child._logger, m_logger)

    def test_set_active(self):
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        mock_children = {n: mock.Mock(spec_set=c, wraps=c)
                         for n, c in dict.items(DUT.children)}
        dict.update(DUT.children, mock_children)

        self.assertTrue(DUT.is_active(), "'active' should be True initially")
        # TODO PY3 check if we can get rid of the list() call
        for child_name, child in list(DUT.children.items()):
            self.assertTrue(child.is_active(),
                            "Child {} should be active".format(child_name))

        # Now set active to false
        DUT.set_active(False)
        self.assertFalse(DUT.is_active(),
                         "'active' should be False after set_active(False)")

        # TODO PY3 check if we can get rid of the list() call
        for child_name, child in list(DUT.children.items()):
            self.assertFalse(child.is_active(),
                            "Child {} should not be active".format(child_name))

        # And now back to to active
        DUT.set_active(True)
        self.assertTrue(DUT.is_active(),
                        "'active' should be True after set_active(True)")
        # TODO PY3 check if we can get rid of the list() call
        for child_name, child in list(DUT.children.items()):
            self.assertTrue(child.is_active(),
                            "Child {} should be active".format(child_name))

    def test_until_sync_states(self):
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        # All children should be in 'disconnected' state, so until_synced() should return
        # an unresolved future and until_not_synced() a resolved future
        self.assertFalse(DUT.until_synced().done())
        self.assertTrue(DUT.until_not_synced().done())

        # Set all child states sync functions to resolved and not_synced to unresolved
        # TODO PY3 check if we can get rid of the list() call
        for child in list(DUT.children.values()):
            f = tornado.concurrent.Future()
            f.set_result(None)
            child.until_synced = mock.create_autospec(
                child.until_synced, return_value=f)
            child.until_not_synced = mock.create_autospec(
                child.until_not_synced, return_value=tornado.concurrent.Future())

        # Now until_synced() should be resolved and until_not_synced() unresolved
        self.assertTrue(DUT.until_synced().done())
        self.assertFalse(DUT.until_not_synced().done())

        # Set only _one_ of the children to not-synced, should be the same as if all of
        # them are disconnected
        for i, child in enumerate(DUT.children.values()):
            if i == 1:
                # Set child to not synced
                f = tornado.concurrent.Future()
                f.set_result(None)
                child.until_not_synced.return_value = f
                child.until_synced.return_value = tornado.concurrent.Future()
            else:
                f = tornado.concurrent.Future()
                f.set_result(None)
                child.until_synced.return_value = tornado.concurrent.Future()
                child.until_not_synced.return_value = f

        self.assertFalse(DUT.until_synced().done())
        self.assertTrue(DUT.until_not_synced().done())


    def test_set_ioloop(self):
        # Make two tornado IOLoop instances, one that is installed as the current thread
        # IOLoop, and one that we will explicity pass to set_ioloop. If set_ioloop is not
        # doing it's job, the children would automatically use thread_ioloop instance.
        thread_ioloop = tornado.ioloop.IOLoop()
        self.addCleanup(thread_ioloop.close, all_fds=True)
        thread_ioloop.make_current()
        our_ioloop = tornado.ioloop.IOLoop()
        self.addCleanup(our_ioloop.close, all_fds=True)
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        DUT.set_ioloop(our_ioloop)
        DUT.start()
        for child_name in self.default_spec_orig['clients']:
            self.assertIs(DUT.children[resource.escape_name(child_name)].ioloop,
                          our_ioloop)


class test_KATCPClientResourceContainerIntegrated(tornado.testing.AsyncTestCase):
    def setUp(self):
        super(test_KATCPClientResourceContainerIntegrated, self).setUp()
        self.default_spec = dict(clients={
            'resource1' : dict(controlled=True),
            'resource2' : dict(controlled=True),
            'resource3' : dict(controlled=True)},
                                 name='intgtest')
        self.resource_names = list(self.default_spec['clients'].keys())
        self.servers = {rn: DeviceTestServer('', 0) for rn in self.resource_names}
        for i, (s_name, s) in enumerate(sorted(self.servers.items())):
            start_thread_with_cleanup(self, s)
            self.default_spec['clients'][s_name]['address'] = s.bind_address
            # Add a unique sensor to each server
            sensor = DeviceTestSensor(DeviceTestSensor.INTEGER, "int."+s_name,
                                      "An Integer.",
                                      "count", [-50, 50], timestamp=self.io_loop.time(),
                                      status=DeviceTestSensor.NOMINAL, value=i)
            s.add_sensor(sensor)
            # Add a unique request to each server
            def handler(self, req, msg):
                """A new command."""
                return Message.reply(msg.name, "ok", "bling1", "bling2")
            s._request_handlers['sparkling-new-'+s_name] = handler

    @tornado.testing.gen_test
    def test_preset_protocol_flags(self):
        pf1 = ProtocolFlags(4, 0, '')
        pf2 = ProtocolFlags(5, 0, 'M')
        pf3 = ProtocolFlags(5, 0, 'I')
        clients = self.default_spec['clients']
        clients['resource1']['preset_protocol_flags'] = pf1
        clients['resource2']['preset_protocol_flags'] = pf2
        clients['resource3']['preset_protocol_flags'] = pf3
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        DUT.start()
        children = DUT.children
        self.assertEqual(
            children.resource1._inspecting_client.katcp_client.protocol_flags,
            pf1)
        self.assertEqual(
            children.resource2._inspecting_client.katcp_client.protocol_flags,
            pf2)
        self.assertEqual(
            children.resource3._inspecting_client.katcp_client.protocol_flags,
            pf3)

    @tornado.testing.gen_test
    def test_timeout_of_until_synced(self):
        self.default_spec_orig = copy.deepcopy(self.default_spec)
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        DUT.start()
        # Test for timing out
        with self.assertRaises(tornado.gen.TimeoutError):
            yield DUT.until_synced(timeout=0.001)
        # Test for NOT timing out
        yield DUT.until_synced(timeout=0.5)

    @tornado.testing.gen_test(timeout=1000)
    def test_set_sampling(self):
        self.default_spec_orig = copy.deepcopy(self.default_spec)
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        DUT.start()

        def side_effect(*args, **kwargs):
            f = tornado.concurrent.futures.Future()
            f.set_result(None)
            return f

        additional = {'resource1': 'sensor_1',
                      'resource2': 'agg_sensor,sensor_1',
                      'resource3': 'sensor_3'}
        for x in additional:
            s = self.servers[x]
            for sens in additional[x].split(","):
                sensor = DeviceTestSensor(DeviceTestSensor.INTEGER, sens,
                                      "An Integer.",
                                      "count", [-50, 50], timestamp=self.io_loop.time(),
                                      status=DeviceTestSensor.NOMINAL, value=0)
                s.add_sensor(sensor)

        yield DUT.until_synced()

        DUT.children.resource1.set_sampling_strategy = mock.Mock(side_effect=side_effect)
        DUT.children.resource2.set_sampling_strategy = mock.Mock(side_effect=side_effect)
        DUT.children.resource3.set_sampling_strategy = mock.Mock(side_effect=side_effect)

        strat1 = ('period', '2.1')
        strat2 = ('event',)
        strat3 = ('event-rate', '2', '3')
        yield DUT.set_sampling_strategy('resource1_sensor_1', strat1)
        DUT.children.resource1.set_sampling_strategy.assert_called_once_with(
            'sensor_1', strat1)
        DUT.children.resource2.set_sampling_strategy.assert_not_called()
        DUT.children.resource3.set_sampling_strategy.assert_not_called()
        DUT.children.resource1.set_sampling_strategy.reset_mock()

        yield DUT.set_sampling_strategy('resource2.sensor_1', strat2)
        DUT.children.resource2.set_sampling_strategy.assert_called_once_with(
            'sensor_1', strat2)
        DUT.children.resource1.set_sampling_strategy.assert_not_called()
        DUT.children.resource3.set_sampling_strategy.assert_not_called()
        DUT.children.resource2.set_sampling_strategy.reset_mock()

        yield DUT.set_sampling_strategy('agg_sensor', strat1)
        DUT.children.resource2.set_sampling_strategy.assert_called_once_with(
            'agg_sensor', strat1)
        DUT.children.resource1.set_sampling_strategy.assert_not_called()
        DUT.children.resource3.set_sampling_strategy.assert_not_called()
        DUT.children.resource2.set_sampling_strategy.reset_mock()

        yield DUT.set_sampling_strategy('resource3_sensor_3', strat3)
        DUT.children.resource3.set_sampling_strategy.assert_called_once_with(
            'sensor_3', strat3)
        DUT.children.resource1.set_sampling_strategy.assert_not_called()
        DUT.children.resource2.set_sampling_strategy.assert_not_called()
        DUT.children.resource3.set_sampling_strategy.reset_mock()

        yield DUT.set_sampling_strategies('sensor_1', strat2)
        DUT.children.resource1.set_sampling_strategy.assert_called_once_with(
            'sensor_1', strat2)
        DUT.children.resource2.set_sampling_strategy.assert_called_once_with(
            'sensor_1', strat2)
        DUT.children.resource3.set_sampling_strategy.assert_not_called()
        DUT.children.resource1.set_sampling_strategy.reset_mock()
        DUT.children.resource2.set_sampling_strategy.reset_mock()

    @tornado.testing.gen_test(timeout=1000)
    def test_set_sensor_listener(self):

        self.default_spec_orig = copy.deepcopy(self.default_spec)
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        DUT.start()

        def side_effect(*args, **kwargs):
            f = tornado.concurrent.futures.Future()
            f.set_result(None)
            return f

        additional = {'resource1': 'sensor_1',
                      'resource2': 'agg_sensor,sensor_1',
                      'resource3': 'sensor_3'}
        for x in additional:
            s = self.servers[x]
            for sens in additional[x].split(","):
                sensor = DeviceTestSensor(DeviceTestSensor.INTEGER, sens,
                                      "An Integer.",
                                      "count", [-50, 50], timestamp=self.io_loop.time(),
                                      status=DeviceTestSensor.NOMINAL, value=0)
                s.add_sensor(sensor)

        yield DUT.until_synced()

        DUT.children.resource1.set_sensor_listener = mock.Mock(side_effect=side_effect)
        DUT.children.resource2.set_sensor_listener = mock.Mock(side_effect=side_effect)
        DUT.children.resource3.set_sensor_listener = mock.Mock(side_effect=side_effect)

        listener1 = lambda *x : None
        listener2 = lambda *y : None
        listener3 = lambda *z : None

        yield DUT.set_sensor_listener('resource1_sensor_1', listener1)
        DUT.children.resource1.set_sensor_listener.assert_called_once_with(
            'sensor_1', listener1)
        DUT.children.resource2.set_sensor_listener.assert_not_called()
        DUT.children.resource3.set_sensor_listener.assert_not_called()
        DUT.children.resource1.set_sensor_listener.reset_mock()

        yield DUT.set_sensor_listener('resource2_sensor_1', listener2)
        DUT.children.resource2.set_sensor_listener.assert_called_once_with(
            'sensor_1', listener2)
        DUT.children.resource1.set_sensor_listener.assert_not_called()
        DUT.children.resource3.set_sensor_listener.assert_not_called()
        DUT.children.resource2.set_sensor_listener.reset_mock()

        yield DUT.set_sensor_listener('agg_sensor', listener2)
        DUT.children.resource2.set_sensor_listener.assert_called_once_with(
            'agg_sensor', listener2)
        DUT.children.resource1.set_sensor_listener.assert_not_called()
        DUT.children.resource3.set_sensor_listener.assert_not_called()
        DUT.children.resource2.set_sensor_listener.reset_mock()

        yield DUT.set_sensor_listener('resource3.sensor_3', listener3)
        DUT.children.resource3.set_sensor_listener.assert_called_once_with(
            'sensor_3', listener3)
        DUT.children.resource1.set_sensor_listener.assert_not_called()
        DUT.children.resource2.set_sensor_listener.assert_not_called()
        DUT.children.resource3.set_sensor_listener.reset_mock()


    @tornado.testing.gen_test(timeout=1000)
    def test_listening(self):
        spec = dict(name="resource1",
            description='resource for testing',
            address=('testhost', 12345),
            controlled=True)

        ##DUT = resource_client.KATCPClientResource(spec)
        ##DUT.start()
        self.default_spec_orig = copy.deepcopy(self.default_spec)
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        DUT.start()


        # Setup a new sensor with listener and trigger some sensor updates
        new_sensor = DeviceTestSensor(DeviceTestSensor.INTEGER, 'new_sensor',
                                      "An Integer.",
                                      "count", [-50, 50], timestamp=self.io_loop.time(),
                                      status=DeviceTestSensor.NOMINAL, value=0)
        server = self.servers["resource1"]
        server.add_sensor(new_sensor)
        yield DUT.until_synced()
        katcp_sensor = DUT.children.resource1.sensor.new_sensor
        import time
        t0 = time.time() - 5 # A little old
        t, status, value = (t0, Sensor.WARN, -1)
        katcp_sensor.set(t, status, value)

        listener = mock.Mock()
        yield DUT.children.resource1.set_sensor_listener('new_sensor', listener)
        t, status, value = (t0+0.5, Sensor.NOMINAL, -2)
        katcp_sensor.set(t, status, value)
        self.assertEqual(katcp_sensor.reading.status, 'nominal')
        self.assertEqual(katcp_sensor.reading.value, -2)
        katcp_sensor.set(t, status, value)
        t, status, value = (t0+1, Sensor.NOMINAL, 10)
        katcp_sensor.set(t, status, value)
        t, status, value = (t0+1.5, Sensor.NOMINAL, 5)
        katcp_sensor.set(t, status, value)
        t, status, value = (t0+2, Sensor.NOMINAL, 4)
        katcp_sensor.set(t, status, value)
        t, status, value = (t0+2.5, Sensor.NOMINAL, 3)
        katcp_sensor.set(t, status, value)
        print("listener.mock_calls", listener.mock_calls)
        self.assertEqual(len(listener.mock_calls), 6)
        return

    def get_expected(self, testserv_attr):
        expected_items = []
        for i, (serv_name, serv) in enumerate(sorted(self.servers.items())):
            for item_name in getattr(serv, testserv_attr):
                expected_items.append((serv_name+'_'+item_name)
                                        .replace('.', '_')
                                        .replace('-', '_'))
        return expected_items

    @tornado.gen.coroutine
    def get_DUT_synced(self):
        # make a copy in case the test or DUT messes up any of the original dicts.
        self.default_spec_orig = copy.deepcopy(self.default_spec)
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        DUT.start()
        yield DUT.until_synced()
        raise tornado.gen.Return(DUT)


    @tornado.testing.gen_test(timeout=1)
    def test_sensors(self):
        DUT = yield self.get_DUT_synced()
        expected_sensors = self.get_expected('sensor_names')
        self.assertEqual(sorted(DUT.sensor), sorted(expected_sensors))
        # Test that some sensor objects are correctly mapped between container and client
        self.assertIs(DUT.sensor.resource1_int_resource1,
                      DUT.children['resource1'].sensor.int_resource1)
        self.assertIs(DUT.sensor.resource2_int_resource2,
                      DUT.children['resource2'].sensor.int_resource2)
        self.assertIs(DUT.sensor.resource3_an_int,
                      DUT.children['resource3'].sensor.an_int)


    @tornado.testing.gen_test(timeout=1)
    def test_requests(self):
        r2_spec = self.default_spec['clients']['resource2']
        r2_spec['always_allowed_requests'] = ['sparkling-new-resource2']
        r2_spec['controlled'] = False
        DUT = yield self.get_DUT_synced()
        # Strip out all resource2 requests (since it is not controlled) except for
        # sparkling-new-resource2 which is in always_allowed_requests.
        expected_requests = [r for r in self.get_expected('request_names')
                             if (not r.startswith('resource2_') or
                                 r == 'resource2_sparkling_new_resource2')]
        self.assertEqual(sorted(DUT.req), sorted(expected_requests))
        # Test that some request objects are correctly mapped between container and client
        self.assertIs(DUT.req.resource1_sparkling_new_resource1,
                      DUT.children['resource1'].req.sparkling_new_resource1)
        self.assertIs(DUT.req.resource2_sparkling_new_resource2,
                      DUT.children['resource2'].req.sparkling_new_resource2)
        self.assertIs(DUT.req.resource3_halt,
                      DUT.children['resource3'].req.halt)

    @tornado.testing.gen_test(timeout=1)
    def test_group_wait(self):
        # Start new container
        self.default_spec_orig = copy.deepcopy(self.default_spec)
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        DUT.add_group('test', list(DUT.children.keys()))
        DUT.start()
        # TODO PY3 check if we can get rid of the list() call
        for server in list(self.servers.values()):
            # Setup a new sensor on all clients to wait on
            wait_sensor = DeviceTestSensor(DeviceTestSensor.INTEGER,
                                           'wait_sensor', "An Integer.",
                                           "count", [-50, 50], value=0,
                                           timestamp=self.io_loop.time(),
                                           status=DeviceTestSensor.NOMINAL)
            server.add_sensor(wait_sensor)
        yield DUT.until_synced()
        # Ensure strategies are set for wait() to work
        # TODO PY3 check if we can get rid of the list() call
        for client in list(DUT.children.values()):
            client.set_sampling_strategy('wait_sensor', 'event')
        group = DUT.groups['test']
        # Test the no timeout case too for what it's worth
        result = yield group.wait('wait_sensor', 0, timeout=None)
        self.assertTrue(result)
        result = yield group.wait('wait_sensor', 0, timeout=None, quorum=1.0)
        self.assertTrue(result)
        # Check detailed results per client
        # TODO PY3 check if we can get rid of the list() call
        for client in list(DUT.children.values()):
            self.assertTrue(result[client.name])
        result = yield group.wait('wait_sensor', 1, timeout=0.1)
        self.assertFalse(result)
        # TODO PY3 check if we can get rid of the list() call
        for client in list(DUT.children.values()):
            self.assertFalse(result[client.name])
        # Test quorum functionality
        with self.assertRaises(TypeError):
            yield group.wait('wait_sensor', 1, timeout=0.1, quorum=1.1)
        # TODO PY3 check if we can get rid of the list() call
        selected_client = list(DUT.children.keys())[0]
        self.servers[selected_client].get_sensor('wait_sensor').set_value(1)
        result = yield group.wait('wait_sensor', 1, timeout=0.1, quorum=1)
        self.assertTrue(result)
        # Be warned that quorum == 1.0 is not the same as quorum == 1 ...
        result = yield group.wait('wait_sensor', 1, timeout=0.1, quorum=1.0)
        self.assertFalse(result)
        result = yield group.wait('wait_sensor', 1, timeout=0.1, quorum=0.1)
        self.assertTrue(result)
        # TODO PY3 check if we can get rid of the list() call
        for client in list(DUT.children.values()):
            if client.name == selected_client:
                self.assertTrue(result[client.name])
            else:
                self.assertFalse(result[client.name])
        result = yield group.wait('wait_sensor', 1, timeout=0.1, quorum=2)
        self.assertFalse(result)
        # TODO PY3 check if we can get rid of the list() call
        for client in list(DUT.children.values()):
            if client.name == selected_client:
                self.assertTrue(result[client.name])
            else:
                self.assertFalse(result[client.name])

    @tornado.testing.gen_test
    def test_no_memory_leak_after_usage(self):
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        wr = weakref.ref(DUT)

        DUT.start()
        yield DUT.until_synced()
        DUT.stop()
        yield DUT.until_stopped()

        # clear strong reference and check if object can be garbage collected
        DUT = None
        gc.collect()
        self.assertIsNone(wr())

    @tornado.testing.gen_test
    def test_no_memory_leak_after_until_not_synced(self):
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        wr = weakref.ref(DUT)

        DUT.start()
        yield DUT.until_synced()
        # call, but don't yield on until_not_synced - we are testing if
        # stopping cleans up the futures it is waiting on
        DUT.until_not_synced()
        DUT.stop()
        yield DUT.until_stopped()

        # clear strong reference and check if object can be garbage collected
        DUT = None
        gc.collect()
        self.assertIsNone(wr())

    @unittest.skip('TODO: (AJ) Fix clean exit while syncing')
    @tornado.testing.gen_test
    def test_no_memory_leak_after_until_synced(self):
        DUT = resource_client.KATCPClientResourceContainer(self.default_spec)
        wr = weakref.ref(DUT)

        DUT.start()
        yield DUT.until_synced()
        self.servers['resource1'].mass_inform(Message.inform('interface-changed'))
        yield DUT.until_not_synced()
        # call, but don't yield on until_synced - we are testing if
        # stopping cleans up the futures it is waiting on
        DUT.until_synced()
        # TODO: (AJ)  Additional note:
        # This test is skipped because the fix hasn't been found.
        # The test fails regardless of whether `until_synced` is called.
        # The problem is to do with stopping while sync is in progress - the
        # container keeps references to the children.
        DUT.stop()
        yield DUT.until_stopped()

        # clear strong reference and check if object can be garbage collected
        DUT = None
        gc.collect()
        self.assertIsNone(wr())


class test_AttrMappingProxy(unittest.TestCase):
    def test_wrapping(self):
        test_dict = AttrDict(a=2, b=1)
        class TestWrapper(object):
            def __init__(self, wrappee):
                self.wrappee = wrappee

            def __eq__(self, other):
                return self.wrappee == other.wrappee

        wrapped_dict = resource_client.AttrMappingProxy(test_dict, TestWrapper)
        # Test keys
        # TODO PY3 check if we can get rid of the list() call
        self.assertEqual(list(wrapped_dict.keys()), list(test_dict.keys()))
        # Test key access:
        for key in test_dict:
            self.assertEqual(wrapped_dict[key].wrappee, test_dict[key])
        # Test attribute access
        for key in test_dict:
            self.assertEqual(getattr(wrapped_dict, key).wrappee,
                             getattr(test_dict, key))
        # Test whole dict comparison
        # TODO PY3 check if we can get rid of the list() call
        self.assertEqual(
            wrapped_dict,
            {k : TestWrapper(v) for k, v in list(test_dict.items())})


class test_ThreadSafeKATCPClientResourceWrapper(unittest.TestCase):
    def setUp(self):
        self.server = DeviceTestServer('', 0)
        start_thread_with_cleanup(self, self.server)

        self.ioloop_manager = ioloop_manager.IOLoopManager(managed_default=True)
        self.io_loop = self.ioloop_manager.get_ioloop()
        self.host, self.port = self.server.bind_address
        self.default_resource_spec = dict(
            name='thething',
            address=self.server.bind_address,
            controlled=True)
        self.client_resource = resource_client.KATCPClientResource(
            self.default_resource_spec)
        self.client_resource.set_ioloop(self.io_loop)
        self.io_loop.add_callback(self.client_resource.start)

        self.ioloop_thread_wrapper = ioloop_manager.IOLoopThreadWrapper(self.io_loop)
        start_thread_with_cleanup(self, self.ioloop_manager, start_timeout=1)
        self.ioloop_thread_wrapper.default_timeout = 1

        self.DUT = resource_client.ThreadSafeKATCPClientResourceWrapper(
            self.client_resource, self.ioloop_thread_wrapper)
        self.DUT.until_synced()

    def test_wrapped_timeout(self):
        self.assertEqual(self.client_resource.state, 'synced')
        # Test timeout
        self.ioloop_thread_wrapper.default_timeout = 0.001
        t0 = time.time()
        with self.assertRaises(TimeoutError):
            self.DUT.until_state('disconnected')
        self.assertLess(time.time() - t0, 0.2)
        # Now make sure we can actualy still wait on the state
        self.ioloop_thread_wrapper.default_timeout = 1
        self.server.stop()
        self.server.join()
        self.DUT.until_state('disconnected')
        self.assertEqual(self.client_resource.state, 'disconnected')
        self.server.start()
        self.DUT.until_state('synced')
        self.assertEqual(self.client_resource.state, 'synced')

    def test_request(self):
        reply = self.DUT.req.sensor_value('an.int')
        last_server_msg = self.server.messages[-1]
        self.assertTrue(reply.succeeded)
        self.assertEqual(str(last_server_msg),
                         '?sensor-value[{}] an.int'.format(int(reply.reply.mid)))

    def test_sensor(self):
        server_sensor = self.server.get_sensor('an.int')
        reading = self.DUT.sensor.an_int.get_reading()
        self.assertEqual(reading.value, server_sensor.read().value)
        server_sensor.set_value(server_sensor.read().value + 5)
        reading = self.DUT.sensor.an_int.get_reading()
        self.assertEqual(reading.value, server_sensor.read().value)


class test_ThreadSafeKATCPClientResourceWrapper_container(unittest.TestCase):

    def setUp(self):
        self.ioloop_manager = ioloop_manager.IOLoopManager(managed_default=True)
        self.io_loop = self.ioloop_manager.get_ioloop()
        self.io_loop.make_current()

        self.ioloop_thread_wrapper = ioloop_manager.IOLoopThreadWrapper(self.io_loop)
        start_thread_with_cleanup(self, self.ioloop_manager, start_timeout=1)
        self.ioloop_thread_wrapper.default_timeout = 1

        self.default_spec = dict(clients={
            'resource1' : dict(controlled=True),
            'resource2' : dict(controlled=True)},
                                 name='wraptest')
        # TODO PY3 check if we can get rid of the list() call
        self.resource_names = list(self.default_spec['clients'].keys())
        self.servers = {rn: DeviceTestServer('', 0) for rn in self.resource_names}
        for i, (s_name, s) in enumerate(sorted(self.servers.items())):
            start_thread_with_cleanup(self, s)
            self.default_spec['clients'][s_name]['address'] = s.bind_address
            # Add a unique sensor to each server
            sensor = DeviceTestSensor(DeviceTestSensor.INTEGER, "int."+s_name,
                                      "An Integer.",
                                      "count", [-50, 50], timestamp=self.io_loop.time(),
                                      status=DeviceTestSensor.NOMINAL, value=i)
            s.add_sensor(sensor)
            # Add a unique request to each server
            def handler(self, req, msg):
                """A new command."""
                return Message.reply(msg.name, "ok", "bling1", "bling2")
            s._request_handlers['sparkling-new-'+s_name] = handler

        self.resource_container = resource_client.KATCPClientResourceContainer(
            self.default_spec)
        self.DUT = resource_client.ThreadSafeKATCPClientResourceWrapper(
            self.resource_container, self.ioloop_thread_wrapper)
        self.DUT.start()
        self.DUT.until_synced()

    def test_sensor(self):
        self.assertEqual(self.DUT.sensor.resource1_int_resource1,
                         self.DUT.children['resource1'].sensor.int_resource1)
        self.assertIs(self.DUT.sensor.resource1_int_resource1.reading,
                      self.resource_container.sensor.resource1_int_resource1.reading)
        self.servers['resource2'].get_sensor('int.resource2').set_value(17)
        reading = self.DUT.sensor.resource2_int_resource2.get_reading()
        self.assertEqual(reading.value, 17)
        self.assertEqual(reading.status, Sensor.STATUSES[Sensor.NOMINAL])
        self.servers['resource2'].get_sensor('int.resource2').set_value(14)
        self.assertEqual(self.DUT.sensor.resource2_int_resource2.get_value(), 14)
        self.servers['resource2'].get_sensor('int.resource2').set_value(
            10, Sensor.WARN)
        self.assertEqual(self.DUT.sensor.resource2_int_resource2.get_status(),
                         Sensor.STATUSES[Sensor.WARN])
        self.assertEqual(self.DUT.sensor.resource2_int_resource2.value, 10)

    def test_children(self):
        self.assertIs(type(self.DUT.children['resource1']),
                      resource_client.ThreadSafeKATCPClientResourceWrapper)
        self.assertIs(self.DUT.children['resource1'].__subject__,
                      self.resource_container.children['resource1'])

        self.assertIs(type(self.DUT.children['resource2']),
                      resource_client.ThreadSafeKATCPClientResourceWrapper)
        self.assertIs(self.DUT.children['resource2'].__subject__,
                      self.resource_container.children['resource2'])


class test_monitor_resource_sync_state(tornado.testing.AsyncTestCase):
    @tornado.testing.gen_test
    def test_monitor_resource_sync_state(self):
        m_res = mock.Mock()
        callback = mock.Mock()
        exit_event = AsyncEvent()
        synced = AsyncEvent()
        not_synced = AsyncEvent()
        m_res.until_synced = synced.until_set
        m_res.until_not_synced = not_synced.until_set
        def set_synced(sync):
            if sync:
                not_synced.clear()
                synced.set()
            else:
                synced.clear()
                not_synced.set()
        loop_done_future = resource_client.monitor_resource_sync_state(
            m_res, callback, exit_event)
        yield tornado.gen.moment
        self.assertEqual(callback.call_args_list, [mock.call(False)])
        callback.reset_mock()
        # Check that it exits if exit_event is set
        exit_event.set()
        yield tornado.gen.moment
        self.assertFalse(callback.called,
                         'No callback should be made when exit_event is set')
        self.assertTrue(loop_done_future.done(),
                        'Monitor loop should terminate when exit_event is set')
        exit_event.clear()
        loop_done_future = resource_client.monitor_resource_sync_state(
            m_res, callback, exit_event)
        set_synced(True)
        yield tornado.gen.moment
        self.assertEqual(callback.call_args_list, [mock.call(False), mock.call(True)])
        callback.reset_mock()
        set_synced(False)
        yield tornado.gen.moment
        self.assertEqual(callback.call_args_list, [mock.call(False)])
        callback.reset_mock()
        # Now check exit_event when synced is set
        set_synced(True)
        yield tornado.gen.moment
        self.assertEqual(callback.call_args_list, [mock.call(True)])
        callback.reset_mock()
        self.assertFalse(loop_done_future.done(),
                        'Monitor loop should only terminate is exit_event is set')
        exit_event.set()
        yield tornado.gen.moment
        self.assertFalse(callback.called)
        self.assertTrue(loop_done_future.done(),
                        'Monitor loop should terminate when exit_event is set')
